"""
Batch normalization, applies a transformation that maintains the mean
layer output close to 0 and the standard deviation close to 1.


Reference:
---------
Ioffe, Sergey, and Christian Szegedy.
Batch normalization: Accelerating deep network training by reducing
internal covariate shift." International conference on machine learning.
pmlr, 2015.

"""

import numpy as np


class Batch_Normalization:
    """
    Class for Batch Normalization
    """

    def __init__(self, layer, momentum=0.99, epsilon=0.001):
        """
        Initialize a BatchNormalization object.

        Parameters
        ----------
        layer : EasySurrogate Layer object
            The layer of a neural network to which Batch Normalization
            must be applied.
        momentum : float, optional
            Parameter of the moving average, applied to the batch mean and
            variance. The default is 0.99.
        epsilon : float, optional
            Small parameter added to the batch variance, to prevent divide
            by zero during normalization. The default is 0.001.

        Returns
        -------
        None.

        """
        self.layer = layer
        self.epsilon = epsilon
        self.momentum = momentum
        self.training = False
        self.mean = 0
        self.var = 1
        self.std = 1
        self.moving_mean = 0
        self.moving_std = 1
        self.init_weights()

    def init_weights(self):
        """
        Initialize the weights used in Batch Normalization, namely
        beta and gamma and associated mometum & squared gradient matrices.

        Beta is in itialized to zero, and gamma to one.

        Returns
        -------
        None.

        """
        # the bias of the normalized pre-activation output.
        self.beta = np.zeros([self.layer.n_neurons, 1])
        # the scaling factor of the normalized pre-activation output.
        self.gamma = np.ones([self.layer.n_neurons, 1])

        # momentum beta
        self.V_beta = np.zeros(self.beta.shape)
        # squared gradient beta
        self.A_beta = np.zeros(self.beta.shape)
        # momentum gamma
        self.V_gamma = np.zeros(self.gamma.shape)
        # squared gradient gamma
        self.A_gamma = np.zeros(self.gamma.shape)

    def set_training(self, training):
        """
        Set the training flag of batch normalization. If true, the pre-activation
        is normalized using the batch mean and standard deviation. If False,
        the running averages of the mean and standard deviation computed during
        training are used instead.

        Parameters
        ----------
        training : bool
            The training flag.

        Returns
        -------
        None.

        """
        assert isinstance(training, bool), "Training flag must be boolean."
        self.training = training

    def normalize(self, v):
        """
        Normalize the pre-activation output of the layer.

        Parameters
        ----------
        v : array, shape (n_neurons, batch_size)
            The unnormalized pre-activation output of the layer.

        Returns
        -------
        a, array, shape (n_neurons, batch_size)
            The normalized pre-activation output of the layer.

        """
        # the linear preactivation output of the layer
        self.v = v

        # if the training flag is set, use the mini-batches to compute
        # the mean and variance
        if self.training and v.shape[1] > 1:

            # compute mean and variance of v over the mini batch
            self.mean = np.mean(self.v, axis=1, keepdims=True)
            self.var = np.var(self.v, axis=1, keepdims=True) + self.epsilon
            self.std = self.var ** 0.5

            # compute a moving average of the moments for inference
            self.moving_mean = self.moving_mean * self.momentum + self.mean * (1 - self.momentum)
            self.moving_std = self.moving_std * self.momentum + self.std * (1 - self.momentum)

            # standardize v
            self.v_hat = (self.v - self.mean) / self.std
            # linearly parameterize v_hat to allow for full expressive power
            # of activation function
            self.a = self.gamma * self.v_hat + self.beta
        # if the training flag is not set, use the moving averages for
        # normalization
        else:
            # standardize v using the computed moving averages
            self.v_hat = (self.v - self.moving_mean) / self.moving_std
            # linearly parameterize v_hat to allow for full expressive power
            self.a = self.gamma * self.v_hat + self.beta

        return self.a

    def compute_L_grad_W(self):
        """
        Compute the gradient of the loss function with respect to the weight
        of the layer. Propagates the loss through the batch-normalization
        procedure to compute the loss gradient wrt the unnormalized 
        pre-activation output v. This gradient is used to compute dL/dW.
        The loss gradients wrt beta and gamma are also computed here.

        Returns
        -------
        L_grad_W : array, shape (n_neurons + n_bias previous layer, n_neurons this layer)
            The loss gradient.

        """

        # keyword arguments to use in all np.sum() calls
        kwargs = {'axis': 1, 'keepdims': True}

        # the gradient loss wrt a, the pre activation output after
        # the batch norm mode
        self.dL_da = self.layer.delta_ho * self.layer.grad_Phi

        # the gradient loss of the parameters of a
        self.L_grad_gamma = np.sum(self.dL_da * self.v_hat, **kwargs)
        self.L_grad_beta = np.sum(self.dL_da, **kwargs)

        # the gradient loss wrt the mini batch variance
        v_minus_mean = self.v - self.mean
        self.dL_dvar = np.sum(v_minus_mean * self.dL_da, **kwargs)
        self.dL_dvar *= -self.gamma / (2 * self.std ** 3)

        # size of the mini batch
        M = self.layer.batch_size

        # the gradient loss wrt the mini batch mean
        self.dL_dmean = self.gamma / self.std ** 3 * \
            np.sum(v_minus_mean * self.dL_da, **kwargs) * \
            np.sum(v_minus_mean / M, **kwargs) - \
            self.gamma / self.std * np.sum(self.dL_da, **kwargs)

        # the gradient loss wrt preactivation output before the
        # batch norm node
        self.dL_dv = self.gamma / self.std * self.dL_da + self.dL_dmean / M + \
            2 * v_minus_mean * self.dL_dvar / M

        # computer the loss gradient wrt the weights of this layer
        h_rm1 = self.layer.layer_rm1.h
        L_grad_W = np.dot(h_rm1, self.dL_dv.T) / M

        return L_grad_W



import numpy as np

from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import DotProduct, WhiteKernel, Matern, RBF, ConstantKernel
from sklearn.metrics import mean_squared_error as mse

import mogp_emulator as mogp
from mogp_emulator import GaussianProcess, MultiOutputGP
from mogp_emulator.MeanFunction import Coefficient, LinearMean, MeanFunction


class GP:

    def __init__(
            self,
            kernel='Matern',
            n_in=1,
            n_out=1,
            length_scale=1.0,
            prefactor=True,
            bias=False,
            noize=1e-8,
            n_iter=1,
            save=True,
            load=False,
            name='GP',
            on_gpu=False,
            backend='scikit-learn',
            standardize_X=True,
            standardize_y=True,
            **kwargs):

        self.n_in = n_in
        self.n_out = n_out

        self.backend = backend

        self.on_gpu = on_gpu

        self.kernel_argument = ''
        self.noize_argument = ''

        # sciki-learn specific part
        if self.backend == 'scikit-learn':
            self.kernel = ConstantKernel(constant_value=1.0,
                                         constant_value_bounds=(1e-6, 1e+6))

            if kernel == 'Matern':
                self.kernel *= Matern(length_scale=[length_scale] *
                                      self.n_in, length_scale_bounds=[length_scale *
                                                                      1e-4, length_scale *
                                                                      1e+4], nu=2.5)

            elif kernel == 'RBF':
                self.kernel *= RBF(length_scale=[length_scale] * self.n_in,
                                   length_scale_bounds=[length_scale * 1e-4, length_scale * 1e+4])

            if bias:
                self.kernel += ConstantKernel(constant_value=1.0,
                                              constant_value_bounds=(1e-5, 1e+5))

            noize_val = 1e-8
            bounds_val = (noize_val * 1e-3, noize_val * 1e+3)
            if noize == 'adaptive':
                noize_val = 1e-12
                bounds_val = 'fixed'
            elif isinstance(noize, float):
                noize_val = noize
                bounds_val = (noize_val * 1e-3, noize_val * 1e+3)

            if noize is not False:
                self.kernel += WhiteKernel(noise_level=noize_val,
                                           noise_level_bounds=bounds_val)

            self.n_iter = n_iter

        # MOGP specific part
        elif self.backend == 'mogp':

            if kernel == 'Matern':
                self.kernel_argument += 'Matern52'
            elif kernel == 'RBF':
                self.kernel_argument += 'SquaredExponential'

            if isinstance(noize, float):
                self.noize_argument = noize
            elif noize == 'fit' or noize is True:
                self.noize_argument = 'fit'
            elif noize == 'adaptive':
                self.noize_argument = 'adaptive'
            else:
                self.noize_argument = 'adaptive'  # redundant, but keeping a default option for future

            if bias:
                raise NotImplementedError('Non-stationary kernels are not implemented in MOGP')

            # self.mean = Coefficient() + Coefficient() * LinearMean()

        else:
            raise NotImplementedError('Currently supporting only scikit-learn and mogp backend')

    def train(self, X, y):

        self.n_train = X.shape[0]

        try:
            n_in = X.shape[1]
            if self.n_in != n_in:
                raise RuntimeError('Size of training data feature is different from expected')
        except IndexError:
            if self.n_in != 1:
                raise RuntimeError(
                    'Size of training data feature is different from expected default =1')

        try:
            n_out = y.shape[1]
            if self.n_out != n_out:
                raise RuntimeError('Size of training data target is different from expected')
        except IndexError:
            if self.n_out != 1:
                raise RuntimeError(
                    'Size of training data target is different from expected default =1')

        if self.backend == 'scikit-learn':
            self.instance = GaussianProcessRegressor(
                kernel=self.kernel, n_restarts_optimizer=self.n_iter, normalize_y=True)
            self.instance.fit(X, y)
            self.kernel = self.instance.kernel_
        elif self.backend == 'mogp':
            if self.n_out == 1:
                self.instance = GaussianProcess(X, y.reshape(-1), kernel=self.kernel_argument,
                                                nugget=self.noize_argument)
            else:
                self.instance = MultiOutputGP(X, y.T, kernel=self.kernel_argument,
                                              nugget=self.noize_argument)
            self.instance = mogp.fit_GP_MAP(self.instance)

    def predict(self, X_i):

        if self.backend == 'scikit-learn':
            # for single sample X_i should be nparray(1, n_feat)
            m, v = self.instance.predict(X_i.reshape(1, -1), return_std=True)
            d = np.zeros(m.shape)
        elif self.backend == 'mogp':
            m, v, d = self.instance.predict(X_i, unc=True, deriv=True)
        else:
            raise NotImplementedError('Non-stationary kernels are not implemented in MOGP')

        return m, v, d

    def forward(self, X_i):  # no cases when required different from .predict() for GP case
        m, v = self.instance.predict(X_i)
        return m  # for single sample should be nparray(1,n_feat)

    def print_model_info(self):
        print('===============================')
        print('Gaussian Process parameters')
        print('===============================')
        if self.backend == 'scikit-learn':
            # print('Kernel params =', self.instance.kernel_.get_params())
            print('Kernel =', self.instance.kernel_)
            print('Kernel theta =', self.instance.kernel_.theta)
            print('Surrogate model parameters :', self.instance.get_params())
        elif self.backend == 'mogp':
            if isinstance(self.instance, MultiOutputGP):
                print('Model parameter printed not implemented for MOGP vector QoI models')
            else:
                print('Kernel =', self.instance.kernel)
                print('Kernel theta =', self.instance.theta)
        print('Output dimensionality =', self.n_out)
        print('Input dimensionality =', self.n_in)
        print('On GPU =', self.on_gpu)
        print('===============================')



"""
==============================================================================
CLASS FOR A DEEP ACTIVE SUBSPACE SURROGATE
------------------------------------------------------------------------------
Author:
        W. Edeling
Method:
        Tripathy, Rohit, and Ilias Bilionis. "Deep active subspaces: A scalable
        method for high-dimensional uncertainty propagation."
        ASME 2019 International Design Engineering Technical Conferences and
        Computers and Information in Engineering Conference.
        American Society of Mechanical Engineers Digital Collection, 2019.
==============================================================================
"""

import numpy as np
import easysurrogate as es
from ..campaign import Campaign


class DAS_Surrogate(Campaign):
    """
    ANN_surrogate class for a standard feed forward neural network.
    """

    def __init__(self, **kwargs):
        print('Creating DAS_Surrogate Object')
        self.name = 'DAS Surrogate'
        # create a Feature_Engineering object
        self.feat_eng = es.methods.Feature_Engineering()

    ############################
    # START COMMON SUBROUTINES #
    ############################

    def train(self, feats, target, d, n_iter, test_frac=0.0,
              n_layers=2, n_neurons=100,
              activation='tanh', activation_das='linear', loss='squared',
              batch_size=64, batch_norm=False,  lamb=0.0,
              standardize_X=True, standardize_y=True, **kwargs):
        """
        Perform backpropagation to train the DAS network

        Parameters
        ----------
        feats : array [n_samples, n_feats]
            The feature array.
        target : array
            The target data.
        d : integer
            The dimension of the active subspace.
        n_iter : integer
            The number of mini batch iterations.
        test_frac : float, optional
            Fraction of the data to use for testing. The default is 0.0.
        n_layers : integer, optional
            The number of hidden layers + the output layer. The default is 2.
        n_neurons : integer, optional
            The number of neurons per hidden layer. The default is 100.
        activation : string, optional
            The activation function to use. The default is 'tanh'. Other options include 'relu',
            'hard_tanh', 'leaky_relu', 'sigmoid', 'softplus'.
        activation_das : string, optional
            The activation function to use in the DAS layer. The default is 'linear'.
        loss : string, optional
            The loss function. The default is 'squared'. Other options include 'cross_entropy'
            and 'logistic'.
        batch_size : integer, optional
            The minibatch size. The default is 64.
        batch_norm : boolean or list of booleans, optional
            Use batch normalization in hidden layers. Not used in the DAS layer.
            The default is False.
        lamb : float, optional
            L2 weight regularization parameter. The default is 0.0.
        standardize_X : Boolean, optional
            Standardize the features. The default is True.
        standardize_y : Boolean, optional
            Standardize the output. The default is True.

        Returns
        -------
        None.

        """

        # test fraction
        self.test_frac = test_frac

        # The dimenions of the active subspace
        self.d = d

        # loss function
        self.loss = loss

        # prepare the training data
        X_train, y_train, _, _ = self.feat_eng.get_training_data(
            feats, target, test_frac=test_frac, train_first=True)

        n_out = y_train.shape[1]

        if loss == 'cross_entropy':
            n_softmax = 1
        else:
            n_softmax = 0

        self.batch_norm = [False, False]   # no batch norm in input and DAS layer
        for i in range(n_layers - 2):
            self.batch_norm.append(batch_norm)
        self.batch_norm.append(False)

        # create the feed-forward ANN
        self.neural_net = es.methods.DAS_network(
            X_train,
            y_train,
            d,
            n_layers=n_layers,
            n_neurons=n_neurons,
            n_out=n_out,
            loss=loss,
            n_softmax=n_softmax,
            activation=activation,
            activation_das=activation_das,
            batch_size=batch_size,
            batch_norm=self.batch_norm,
            lamb=lamb,
            decay_step=10**4,
            decay_rate=0.9,
            standardize_X=standardize_X,
            standardize_y=standardize_y,
            save=False,
            **kwargs)

        print('===============================')
        print('Training Deep Active Subspace Neural Network...')

        # set the training flag to True in any layer that uses batch normalization
        self.neural_net.set_batch_norm_training_flag(True)

        # train network for n_iter mini batches
        self.neural_net.train(n_iter, store_loss=True)
        self.set_data_stats()

        # set the training flag to False in any layer that uses batch normalization
        self.neural_net.set_batch_norm_training_flag(False)

    def derivative(self, x, norm=True):
        """
        Compute a derivative of the network output f(x) with respect to the inputs x.

        Parameters
        ----------
        x : array
            A single feature vector of shape (n_in,) or (n_in, 1), where n_in is the
            number of input neurons.
        norm : Boolean, optional, default is True
            Compute the gradient of ||f||_2. If False it computes the gradient of
            f, if f is a scalar. If False and f is a vector, the resulting gradient is of the
            column sum of the full Jacobian matrix.

        Returns
        -------
        df_dx : array
            The derivatives [d||f||_2/dx_1, ..., d||f||_2/dx_n_in]

        """
        # check that x is of shape (n_in, ) or (n_in, 1)
        assert x.shape[0] == self.neural_net.n_in, \
            "x must be of shape (n_in,): %d != %d" % (x.shape[0], self.neural_net.n_in)

        if x.ndim > 1:
            assert x.shape[1] == 1, "Only pass 1 feature vector at a time"

        # set the batch size to 1 if not done already
        if not self.neural_net.batch_size == 1:
            self.neural_net.set_batch_size(1)

        # standardize the input (if inputs were not standardized, feat_mean=0 and feat_std=1)
        x = (x - self.feat_mean) / self.feat_std

        # feed forward and compute and the derivatives
        df_dx = self.neural_net.d_norm_y_dX(x.reshape([1, -1]), feed_forward=True, norm=norm)

        return df_dx

    def predict(self, feat):
        """
        Make a prediction with the Deep Active Subspace network at input parameter settings
        given by feat.

        Parameters
        ----------
        feat : array
               The feature array; the values of the input parameters.

        Returns
        -------
        array
            The prediction of the DAS neural net.

        """

        # if features were standardized during training, do so here as well
        feat = (feat - self.feat_mean) / self.feat_std
        if self.loss == 'cross_entropy':
            # y = the probability mass function at the output layer
            y, _, _ = self.neural_net.get_softmax(feat.reshape([1, self.neural_net.n_in]))
        else:
            # feed forward prediction step
            y = self.neural_net.feed_forward(feat.reshape([1, self.neural_net.n_in])).flatten()
            # transform y back to physical domain
            y = y * self.output_std + self.output_mean

        return y

    def save_state(self):
        """
        Save the state of the DAS surrogate to a pickle file
        """
        state = self.__dict__
        super().save_state(state=state, name=self.name)

    def load_state(self):
        """
        Load the state of the DAS surrogate from file
        """
        super().load_state(name=self.name)

    ##########################
    # END COMMON SUBROUTINES #
    ##########################

    def set_data_stats(self):
        """
        If the data were standardized, this stores the mean and
        standard deviation of the features in the ANN_Surrogate object.
        """
        if hasattr(self.neural_net, 'X_mean'):
            self.feat_mean = self.neural_net.X_mean
            self.feat_std = self.neural_net.X_std
        else:
            self.feat_mean = 0.0
            self.feat_std = 1.0

        if hasattr(self.neural_net, 'y_mean'):
            self.output_mean = self.neural_net.y_mean
            self.output_std = self.neural_net.y_std
        else:
            self.output_mean = 0.0
            self.output_std = 1.0

    def get_dimensions(self):
        """
        Get some useful dimensions of the DAS surrogate. Returns a dict with the number
        of training samples (n_train), the number of data samples (n_samples),
        the number of test samples (n_test), the total number of input parameters (D),
        the dimension of the active subspace (d), and the size of the ouput quantity of
        interest (n_out).

        Returns
        -------
        dims : dict
            The dimensions dictionary.

        """

        dims = {}
        dims['n_train'] = self.feat_eng.n_train
        dims['n_samples'] = self.feat_eng.n_samples
        dims['n_test'] = dims['n_samples'] - dims['n_train']
        dims['D'] = self.neural_net.n_in
        dims['d'] = self.d
        dims['n_out'] = self.neural_net.n_out

        return dims



"""
Class for an artificial neural network.
"""

import pickle
import numpy as np
from scipy.stats import bernoulli
from tqdm import tqdm

try:
    import tkinter as tk
    from tkinter import filedialog
except ImportError:
    print("WARNING: tkinter not imported")
    print("save_ANN and load_ANN have no graphical support, \
          use these by specifying file_path=")

from .Layer import Layer
from .DAS_Layer import DAS_Layer


class ANN:
    """
    Class for an artificial neural network.
    """

    def __init__(self, X, y, alpha=0.001, decay_rate=1.0, decay_step=10**5, beta1=0.9,
                 beta2=0.999, lamb=0.0, n_out=1, loss='squared', activation='tanh',
                 activation_out='linear', n_softmax=0, n_layers=2, n_neurons=16,
                 bias=True, batch_size=1, batch_norm=False,
                 param_specific_learn_rate=True,
                 save=False, on_gpu=False, name='ANN',
                 standardize_X=True, standardize_y=True, **kwargs):
        """
        Initialize the Artificial Neural Network object.

        Parameters
        ----------
        X : array
            The input features.
        y : array
            The target data.
        alpha : float, optional
            The learning rate. The default is 0.001.
        decay_rate : float, optional
            Factor multiplying the decay rate every decay_step iterations.
            The default is 1.0.
        decay_step : int, optional
            The number of training iterations after which decay_rate is lowered.
            The default is 10**5.
        beta1 : float, optional
            Momentum parameter controlling the moving average of the loss gradient.
            Used for the parameter-specific learning rate. The default is 0.9.
        beta2 : float, optional
            Parameter controlling the moving average of the squared gradient.
            Used for the parameter-specific learning rate. The default is 0.999.
        lamb : float, optional
            L2 weight regularization parameter. The default is 0.0.
        n_out : int, optional
            The number of output neurons. The default is 1.
        loss : string, optional
            The name of the loss function. The default is 'squared'.
        activation : string or list of strings, optional
            The name of the activation function of the hidden layers.
            The default is 'tanh'.
        activation_out : string, optional
            The name of the activation function of the output layer.
            The default is 'linear'.
        n_softmax : int, optional
            The number of softmax layers attached to the output. The default is 0.
        n_layers : integer or list of integers, optional
            The number of layers, not counting the input layer. The default is 2.
        n_neurons : int, optional
            The number of neurons per hidden layer. The default is 16.
        bias : boolean, optional
            Use a bias neuron. The default is True.
        batch_size : int, optional
            The size of the mini batch. The default is 1.
        batch_norm : boolean or list of booleans, optional
            Use batch normalization. The default is False.
        param_specific_learn_rate : boolean, optional
            Use parameter-specific learing rate. The default is True.
        save : boolean, optional
            Save the neural network to a pickle file after training.
            The default is False.
        on_gpu : boolean, optional
            Train the neural network on a GPU using cupy. NOT IMPLEMENTED IN THIS VERSION.
            The default is False.
        name : string, optional
            The name of the neural network. The default is 'ANN'.
        standardize_X : boolean, optional
            Standardize the features. The default is True.
        standardize_y : boolean, optional
            Standardize the target data. The default is True.

        Returns
        -------
        None.

        """

        # the features
        self.X = X

        # number of training data points
        self.n_train = X.shape[0]

        # the training outputs
        self.y = y

        # number of input nodes
        try:
            self.n_in = X.shape[1]
        except IndexError:
            self.n_in = 1

        # number of output neurons
        self.n_out = n_out

        # #use either numpy or cupy via xp based on the on_gpu flag
        # global xp
        # if on_gpu == False:
        #     import numpy as xp
        # else:
        #     import cupy as xp

        # self.on_gpu = on_gpu
        self.on_gpu = False

        # standardize the training data
        if standardize_X:

            self.X_mean = np.mean(X, axis=0)
            self.X_std = np.std(X, axis=0)
            self.X = (X - self.X_mean) / self.X_std

        if standardize_y:
            self.y_mean = np.mean(y, axis=0)
            self.y_std = np.std(y, axis=0)
            self.y = (y - self.y_mean) / self.y_std
        self.standardize_X = standardize_X
        self.standardize_y = standardize_y

        # size of the mini batch used in stochastic gradient descent
        self.batch_size = batch_size

        assert type(batch_norm) is bool or type(batch_norm) is list, \
            "batch_norm must be a boolean value or a list of boolean values"

        # manually specified flags for the use of batch normalization
        if type(batch_norm) is list:
            assert len(batch_norm) == n_layers + 1, \
                "batch_norm flag must be set for every layer"
            self.batch_norm = batch_norm
        # standard flag for batch normalization
        else:
            self.batch_norm = [False]   # no batch norm in input layer
            for i in range(n_layers - 1):
                self.batch_norm.append(batch_norm)
            self.batch_norm.append(False)

        # set dropout to False, can be changed in train subroutine
        self.dropout = False

        ###########################################

        # use a user-specified list of layer objects to create the ANN
        if 'layers' in kwargs:
            # user-specified layers
            self.layers = kwargs['layers']

            assert isinstance(self.layers, list), "layers must be stored in a list"

            # number of layers (hidden + output)
            self.n_layers = len(self.layers) - 1

            # number of neurons in the hidden layers
            self.n_neurons = [layer.n_neurons for layer in self.layers[1:-1]]

            # bias per layer
            self.bias = [layer.bias for layer in self.layers]

            # loss function
            self.loss = self.layers[-1].loss

            # activation function per hidden layer
            self.activation = [layer.activation for layer in self.layers[1:-1]]

            # activation of the output layer
            self.activation_out = self.layers[-1].activation

            # L2 regularization parameter
            self.lamb = self.layers[-1].lamb

            # number of softmax layers at the output
            self.n_softmax = self.layers[-1].n_softmax

            self.set_batch_size(self.batch_size)

        # the default option, create the layers using the parameters
        # of this subroutine
        else:
            self.layers = []

            # number of layers (hidden + output)
            self.n_layers = n_layers

            self.n_neurons = n_neurons

            # use bias neurons
            if isinstance(bias, bool):
                self.bias = []
                for i in range(n_layers):
                    self.bias.append(bias)
            else:
                self.bias = bias

            # loss function type
            self.loss = loss

            # activation function of the hidden layers
            self.activation = activation

            # activation function of the output layer
            self.activation_out = activation_out

            # L2 regularization parameter
            self.lamb = lamb

            # number of sofmax layers
            self.n_softmax = n_softmax

        ############################################

        # bias type checking
        assert isinstance(
            bias, bool) or isinstance(
            bias, list), "bias must be boolean or a list of boolan"

        # type checking the number of neurons
        assert isinstance(self.n_neurons, int) or isinstance(self.n_neurons, list), \
            "n_neurons must be a list or an integer"

        # list of all layer sizes
        self.layer_sizes = [self.n_in]

        # constant size hidden layer
        if isinstance(self.n_neurons, int):
            for i in range(n_layers - 1):
                self.layer_sizes.append(self.n_neurons)
        # variable user-specified size hidden layers
        else:
            for i in range(n_layers - 1):
                self.layer_sizes.append(self.n_neurons[i])

        self.layer_sizes.append(n_out)

        # type checking the activation
        assert isinstance(activation, str) or isinstance(activation, list), \
            "activation must be a string or a list of strings"

        # list of activation functions per layer
        self.layer_activation = ['linear']
        # same activation for each hidden layer
        if isinstance(self.activation, str):
            for i in range(self.n_layers - 1):
                self.layer_activation.append(self.activation)
        else:
            for i in range(self.n_layers - 1):
                self.layer_activation.append(self.activation[i])

        self.layer_activation.append(self.activation_out)

        # training rate
        self.alpha = alpha

        # the rate of decay and decay step for alpha
        self.decay_rate = decay_rate
        self.decay_step = decay_step

        # momentum parameter
        self.beta1 = beta1

        # squared gradient parameter
        self.beta2 = beta2

        # use parameter specific learning rate
        self.param_specific_learn_rate = param_specific_learn_rate

        # save the neural network after training
        self.save = save
        self.name = name

        # additional variables/dicts etc that must be stored in the ann object
        self.aux_vars = kwargs

        self.loss_vals = []

        # initialize network using standard setting if no layers
        # are specified
        if 'layers' not in kwargs:
            self.init_network(**kwargs)

        # connect each layer with its neighbours
        self.connect_layers()

        # print some network stats to screen
        self.print_network_info()

    def init_network(self, **kwargs):
        """
        Set up the network structure by creating the Layer objects and
        connecting them together.

        Returns
        -------
        None.

        """

        # add the inputs and hidden layers
        for r in range(self.n_layers):
            self.layers.append(
                Layer(
                    self.layer_sizes[r],
                    r,
                    self.n_layers,
                    self.layer_activation[r],
                    self.loss,
                    self.bias[r],
                    batch_size=self.batch_size,
                    batch_norm=self.batch_norm[r],
                    lamb=self.lamb,
                    on_gpu=self.on_gpu,
                    **kwargs))

        # add the output layer
        self.layers.append(
            Layer(
                self.n_out,
                self.n_layers,
                self.n_layers,
                self.activation_out,
                self.loss,
                bias=False,
                batch_size=self.batch_size,
                lamb=self.lamb,
                n_softmax=self.n_softmax,
                on_gpu=self.on_gpu,
                **kwargs))

        # self.connect_layers()

    def connect_layers(self):
        """
         Connect each layer in the neural network with its neighbours

        Returns
        -------
        None.

        """

        self.layers[0].meet_the_neighbors(None, self.layers[1])
        self.layers[-1].meet_the_neighbors(self.layers[-2], None)

        for i in range(1, self.n_layers):
            self.layers[i].meet_the_neighbors(self.layers[i - 1], self.layers[i + 1])

    def feed_forward(self, X_i, batch_size=1):
        """
        Run the network forward.

        Parameters
        ----------
        X_i : array
            The feauture array, needs to have shape [batch size, number of features].
        batch_size : int, optional
            The bath size. The default is 1.

        Returns
        -------
        array
            The prediction of the neural network.

        """

        # set the features at the output of in the input layer
        if not self.bias[0]:
            self.layers[0].h = X_i.T
        else:
            self.layers[0].h = np.ones([self.n_in + 1, batch_size])
            self.layers[0].h[0:self.n_in, :] = X_i.T

        # apply dropout to the input layer
        if self.dropout:
            r = bernoulli.rvs(self.dropout_prob[0], size=X_i.T.shape)
            self.layers[0].h[0:self.n_in, :] *= r

        for i in range(1, self.n_layers):
            # compute the output on the layer using matrix-maxtrix multiplication
            if self.dropout:  # with dropout
                self.layers[i].compute_output(batch_size, dropout=self.dropout,
                                              dropout_prob=self.dropout_prob[i])
            else:  # without
                self.layers[i].compute_output(batch_size)

        # output layer, never use dropout here
        self.layers[i + 1].compute_output(batch_size)

        return self.layers[-1].h

    def get_last_prediction(self):
        return self.layers[-1].h

    def get_softmax(self, X_i):
        """
        Get the output of the softmax layer.

        Parameters
        ----------
        X_i : array
            The input features.

        Returns
        -------
        probs : array
            the probabilities of the softmax layer.
        idx_max : int
            The softmax output with the highest probability.

        """
        # feed forward features X_i
        h = self.feed_forward(X_i, batch_size=1)

        probs = []
        idx_max = []
        # rvs = []

        # split the output of the last layer over the number of softmax layers
        for h_i in np.split(h, self.n_softmax):
            # compute the softmax probabilities.
            o_i = np.exp(h_i) / np.sum(np.exp(h_i), axis=0)
            o_i = o_i / np.sum(o_i)
            probs.append(o_i)
            # the softmax output w
            idx_max.append(np.argmax(o_i))

            # draw a random sample from the discrete distribution o_i.
            # TODO: Slow for some reason. Find faster implementation.
            # pmf = rv_discrete(values=(np.arange(o_i.size), o_i.flatten()))
            # rvs.append(pmf.rvs())

        # return values and index of highest probability and random samples from pmf
        return probs, idx_max, None

    def d_norm_y_dX(self, X_i, batch_size=1, feed_forward=True, norm=True, layer_idx=0):
        """
        Compute the derivatives of the squared L2 norm of the output wrt
        the inputs.

        Parameters
        ----------
        X_i : array
            The input features.
        batch_size : int, optional
            The batch size. The default is 1.
        feed_forward: Boolean, optional, default is True.
            Feed X_i forward before computing the gradient.
        norm : Boolean, optional, default is True.
            Compute the gradient of ||y||_2. If False it computes the gradient of
            y, if y is a scalar. If False and y is a vector, the resulting gradient is the
            column sum of the full Jacobian matrix.
        layer_idx : int, optional, default is 0.
            Index for the layer of which to return the derivative. Default is 0, the input layer.

        Returns
        -------
        array
            The derivatives [d||y||_2/dX_1, ..., d||y||_2/dX_n_in]

        """
        if feed_forward:
            self.feed_forward(X_i, batch_size=batch_size)

        for i in range(self.n_layers, -1, -1):
            self.layers[i].compute_delta_hy(norm=norm)
            # also compute the gradient of the output wrt the weights
            if i > 0:
                self.layers[i].compute_y_grad_W()

        # delta_hy of the (input) layer = the derivative of the normed output
        return self.layers[layer_idx].delta_hy

    def back_prop(self, y_i):
        """
        Back-propagation algorithm to find gradient of the loss function with respect
        to the weights of the neural network.

        Parameters
        ----------
        y_i : array
            The target data on which to evaluate the loss funcion.

        Returns
        -------
        None.

        """

        # start back propagation over hidden layers, starting with output layer
        for i in range(self.n_layers, 0, -1):
            self.layers[i].back_prop(y_i)
        self.layers[0].compute_delta_ho()

    def batch(self, X_i, y_i):
        """
        Run a minibatch. Feed X_i forward through the network and
        compute the loss gradient via back propagation.

        Parameters
        ----------
        X_i : array
            The input features of the mini batch.
        y_i : array
            The target data of the mini batch.

        Returns
        -------
        None.

        """

        self.feed_forward(X_i, self.batch_size)
        self.back_prop(y_i)

    def update_weights(self, alpha=0.001, **kwargs):
        """
        Update the weights after a mini batch.

        In the case of a deep-active subspace layer, update the weights of the
        neural network and the weights of the Gram-Schmidt vectors using a
        mini batch.

        Parameters
        ----------
        X_i : array
            The input features of the mini batch.
        y_i : array
            The target data of the mini batch.
        alpha : float, optional
            The learning rate. The default is 0.001.

        Returns
        -------
        None.

        """

        # self.feed_forward(X_i, self.batch_size)
        # self.back_prop(y_i)

        for r in range(1, self.n_layers + 1):

            layer_r = self.layers[r]

            # Deep active subspace layer
            if isinstance(layer_r, DAS_Layer):
                # momentum
                layer_r.V = self.beta1 * layer_r.V + (1.0 - self.beta1) * layer_r.L_grad_Q
                # moving average of squared gradient magnitude
                layer_r.A = self.beta2 * layer_r.A + (1.0 - self.beta2) * layer_r.L_grad_Q**2
            # standard layer
            else:
                # momentum
                layer_r.V = self.beta1 * layer_r.V + (1.0 - self.beta1) * layer_r.L_grad_W
                # moving average of squared gradient magnitude
                layer_r.A = self.beta2 * layer_r.A + (1.0 - self.beta2) * layer_r.L_grad_W**2

                # if batch normalization is used
                if self.batch_norm[r]:
                    # momentum beta parameter
                    layer_r.bn.V_beta = self.beta1 * layer_r.bn.V_beta + \
                        (1.0 - self.beta1) * layer_r.bn.L_grad_beta
                    # moving average of squared gradient magnitude
                    layer_r.bn.A_beta = self.beta2 * layer_r.bn.A_beta + \
                        (1.0 - self.beta2) * layer_r.bn.L_grad_beta**2

                    # momentum beta parameter
                    layer_r.bn.V_gamma = self.beta1 * layer_r.bn.V_gamma + \
                        (1.0 - self.beta1) * layer_r.bn.L_grad_gamma
                    # moving average of squared gradient magnitude
                    layer_r.bn.A_gamma = self.beta2 * layer_r.bn.A_gamma + \
                        (1.0 - self.beta2) * layer_r.bn.L_grad_gamma**2

            # select learning rate
            if not self.param_specific_learn_rate:
                # same learning rate for all weights
                alpha_i = alpha_i_beta = alpha_i_gamma = alpha
            # param specific learning rate
            else:
                # RMSProp
                alpha_i = alpha / (np.sqrt(layer_r.A + 1e-8))

                if self.batch_norm[r]:
                    # RMSProp param specific learning rate for alpha and beta
                    alpha_i_beta = alpha / (np.sqrt(layer_r.bn.A_beta + 1e-8))
                    alpha_i_gamma = alpha / (np.sqrt(layer_r.bn.A_gamma + 1e-8))

                    # update beta and gamma
                    layer_r.bn.beta = layer_r.bn.beta - alpha_i_beta * layer_r.bn.V_beta
                    layer_r.bn.gamma = layer_r.bn.gamma - alpha_i_gamma * layer_r.bn.V_gamma

            # gradient descent update step with L2 regularization
            if self.lamb > 0.0:
                layer_r.W = (1.0 - layer_r.Lamb * alpha_i) * layer_r.W - alpha_i * layer_r.V
            # without regularization
            else:
                # Deep active subspace layer
                if isinstance(layer_r, DAS_Layer):
                    # update the Q weights
                    layer_r.Q = layer_r.Q - alpha_i * layer_r.V
                    # compute the weights W(Q) via Gram Schmidt
                    layer_r.compute_weights()
                # standard layer
                else:
                    layer_r.W = layer_r.W - alpha_i * layer_r.V

    def train(
            self,
            n_batch,
            store_loss=True,
            sequential=False,
            verbose=True,
            dropout=False, **kwargs):
        """
        Train the neural network using stochastic gradient descent.

        Parameters
        ----------
        n_batch : int
            The number of mini-batch iterations.
        store_loss : boolean, optional
            Store the values of the loss function. The default is True.
        sequential : boolean, optional
            Sample a sequential slab of data, starting from a random point.
            The default is False.
        verbose : boolean, optional
            Print information to screen while training. The default is True.
        dropout : boolean, optional
            Use dropout regularization. The default is False. To manually
            specify the dropout probabilities, specify the keyword argument
            "dropout_prob", as a list of probabilities of retaining neurons
            per layer. Otherwise, 0.8 is used for the input layer,
            and 0.5 for the hidden layers.

        Returns
        -------
        None.

        """

        if dropout:
            self.dropout = dropout
            # use standard dropout probabilities
            if 'dropout_prob' not in kwargs:
                self.dropout_prob = [0.8]
                for i in range(self.n_layers - 1):
                    self.dropout_prob.append(0.5)
            # user-specified dropout probabilities
            else:
                self.dropout_prob = kwargs['dropout_prob']

        # set the training flag to True for any layer that uses batch normalization
        self.set_batch_norm_training_flag(True)

        # loop with tqdm progress bar
        for i in tqdm(range(n_batch)):

            # select a random training instance (X, y)
            if not sequential:
                rand_idx = np.random.randint(0, self.n_train, self.batch_size)
            # select a random starting point, and use sequential data from there
            else:
                if self.n_train > self.batch_size:
                    start = np.random.randint(0, self.n_train - self.batch_size, 1)
                else:
                    start = 0
                rand_idx = np.arange(start, start + self.batch_size)

            # compute learning rate
            alpha = self.alpha * self.decay_rate**(int(i / self.decay_step))

            # run the batch
            self.batch(
                self.X[rand_idx],
                self.y[rand_idx].T)

            # update the weights based on the computed loss gradient
            self.update_weights(alpha=alpha, **kwargs)

            # store the loss value
            if store_loss:
                l = self.layers[-1].L_i
                loss_i = np.mean(l)
                self.loss_vals.append(loss_i)

                if np.mod(i, 1000) == 0:
                    if verbose:
                        # print('Batch', i, 'learning rate', alpha, 'loss:', loss_i)
                        tqdm.write(' loss = %.4f' % (loss_i,))

        if self.dropout:
            # scale all weight matrices by dropout prob after training
            for i in range(1, self.n_layers + 1):
                self.layers[1].W *= self.dropout_prob[i - 1]

            # turn off dropout after training
            self.dropout = False

        # set the training flag to False for any layer that uses batch normalization
        self.set_batch_norm_training_flag(False)

        if self.save:
            self.save_ANN()

    def save_ANN(self, file_path="", store_data=False):
        """
        Save the neural network to a picke file.

        Parameters
        ----------
        file_path : string, optional
            The full path of the pickle file. The default is "", in which case a
            filedialog window opens.
        store_data : boolean, optional
            Store the training data as well. The default is False, since storing a
            large amount of data in a pickel file is inefficient.

        Returns
        -------
        None.

        """
        if len(file_path) == 0:

            root = tk.Tk()
            root.withdraw()

            file = filedialog.asksaveasfile(title="Save network",
                                            mode='wb', defaultextension=".pickle")
        else:
            file = open(file_path, 'wb')

        print('Saving ANN to', file.name)

        if store_data:
            # store everything, also data
            pickle.dump(self.__dict__, file)
        else:
            tmp = self.__dict__.copy()
            # do not store data to pickle
            tmp['X'] = []
            tmp['y'] = []
            pickle.dump(tmp, file)

        file.close()

    def load_ANN(self, file_path=""):
        """
        Load the neural network from a pickle file.

        Parameters
        ----------
        file_path : string, optional
            The full path of the pickle file. The default is "", in which case a
            filedialog window opens.

        Returns
        -------
        None.

        """

        # select file via GUI is file_path is not specified
        if len(file_path) == 0:

            root = tk.Tk()
            root.withdraw()

            file_path = filedialog.askopenfilename(title="Open neural network",
                                                   filetypes=(('pickle files', '*.pickle'),
                                                              ('All files', '*.*')))

        print('Loading ANN from', file_path)

        file = open(file_path, 'rb')
        self.__dict__ = pickle.load(file)
        file.close()

        if self.__dict__['X'] == []:
            print('===============================')
            print('**Warning: ANN was saved without training data**')
            print('===============================')

        self.print_network_info()

    def set_batch_size(self, batch_size):
        """
        Set the batch size in all the layers.

        Parameters
        ----------
        batch_size : int
            The batch size.

        Returns
        -------
        None.

        """

        self.batch_size = batch_size

        for i in range(self.n_layers + 1):
            self.layers[i].batch_size = batch_size

    def set_batch_norm_training_flag(self, training):
        """
        Set the training flag for any layers that use batch normalization.
        If True normalization is done with the mean and standard deviation
        computed per mini batch. During inference (training = False), a
        moving average of both moments (computed during training) is used
        for normalization.

        Parameters
        ----------
        training : bool
            Training flag.

        Returns
        -------
        None.

        """

        for r in np.where(self.batch_norm)[0]:
            self.layers[r].bn.set_training(training)

    def compute_misclass_softmax(self, X=None, y=None):
        """
        Compute the number of misclassifications for the sofmax layer(s).

        Parameters
        ----------
        X : array, optional
            Feature array. The default is None, in which case the entire training set is used.
        y : array, optional
            Target data array. The default is None, in which case the entire
            training set is used.

        Returns
        -------
        float
            The misclassification percentage in [0,1] per softmax layer.

        """

        n_misclass = np.zeros(self.n_softmax)

        # compute misclassification error of the training set if X and y are not set
        if y is None:
            print('Computing number of misclassifications wrt all training data.')
            X = self.X
            y = self.y
        else:
            print('Computing number of misclassifications wrt specified data, %d samples' % (y.size,))

        n_samples = X.shape[0]
        error_idx = []

        # loop with tqdm progress bar
        for i in tqdm(range(n_samples)):
            _, max_idx_ann, _ = self.get_softmax(X[i].reshape([1, self.n_in]))

            max_idx_data = np.array([np.where(y_j == 1.0)[0]
                                     for y_j in np.split(y[i], self.n_softmax)])

            for j in range(self.n_softmax):
                if max_idx_ann[j] != max_idx_data[j]:
                    n_misclass[j] += 1
                    error_idx.append(i)

        print('Number of misclassifications =', n_misclass)
        print('Misclassification percentage =', n_misclass / n_samples * 100, '%')

        return n_misclass / n_samples, error_idx

    def get_n_weights(self):
        """
        Return the number of weights

        Returns
        -------
        n_weights : int
            The number of weights.

        """

        n_weights = 0

        for i in range(1, self.n_layers + 1):
            n_weights += self.layers[i].W.size

        print('This neural network has', n_weights, 'weights.')

        return n_weights

    def print_network_info(self):
        """
        Print some characteristics of the neural network to screen.

        Returns
        -------
        None.

        """
        print('===============================')
        print('Neural net parameters')
        print('===============================')
        print('Number of layers =', self.n_layers)
        print('Number of features =', self.n_in)
        print('Loss function =', self.loss)
        print('Number of neurons per hidden layer =', self.n_neurons)
        print('Number of output neurons =', self.n_out)
        print('Activation =', self.layer_activation)
        # print('On GPU =', self.on_gpu)
        self.get_n_weights()
        print('===============================')



"""
CLASS FOR A GAUSSIAN PROCESS REGRESSION
------------------------------------------------------------------------------
Author: Y. Yudin
==============================================================================
"""

import numpy as np
from ..campaign import Campaign
import easysurrogate as es

from sklearn.preprocessing import StandardScaler


class GP_Surrogate(Campaign):

    def __init__(self, backend='scikit-learn', **kwargs):
        """
        GP_surrogate class for Gaussian Process Regression

        """
        print('Creating Gaussian Process Object')

        self.name = 'GP Surrogate'
        self.feat_eng = es.methods.Feature_Engineering()
        self.x_scaler = StandardScaler()
        self.y_scaler = StandardScaler()
        self.backend = backend

        if 'noise' in kwargs:
            self.noise = kwargs['noise']

        if 'n_in' in kwargs:
            self.n_in = kwargs['n_in']
        else:
            self.n_in = 1

        if 'n_out' in kwargs:
            self.n_out = kwargs['n_out']
        else:
            self.n_out = 1

    ############################
    # START COMMON SUBROUTINES #
    ############################

    def train(self, feats, target, n_iter=0,
              test_frac=0.0,
              **kwargs):
        """

        Args:
            feats: feature array, or list of different feature arrays
            target: the target data
            n_iter: number of hyperoptimisation restarts
            test_frac: Fraction of the data used for training

        Returns:
        -------
        None.
        """

        if 'basekernel' not in kwargs:
            self.base_kernel = 'Matern'
        else:
            self.base_kernel = kwargs['basekernel']

        if 'noize' not in kwargs:
            self.noize = 'True'
        else:
            self.noize = kwargs['noize']

        # prepare the training data
        X_train, y_train, X_test, y_test = self.feat_eng.get_training_data(
            feats, target, local=False, test_frac=test_frac, train_first=False)

        # scale the training data
        X_train = self.x_scaler.fit_transform(X_train)
        y_train = self.y_scaler.fit_transform(y_train)
        if len(X_test) > 0 and len(y_test) > 0:
            X_test = self.x_scaler.transform(X_test)
            y_test = self.y_scaler.transform(y_test)

        self.X_train = X_train
        self.y_train = y_train
        self.X_test = X_test
        self.y_test = y_test

        # create a GP process
        print('===============================')
        print('Fitting Gaussian Process...')
        self.model = es.methods.GP(
            kernel=self.base_kernel,
            n_in=self.n_in,
            n_out=self.n_out,
            bias=False,
            noize=self.noize,
            backend=self.backend)

        # get dimensionality of the output
        self.n_out = y_train.shape[1]

        # get the dimesionality of te input
        self.n_in = X_train.shape[1]

        self.model.train(self.X_train, self.y_train)

    def predict(self, X):
        """
        Make a stochastic prediction of the output y conditional on the
        input features [X]

        Args:
            X: list of feature arrays, the state given at the point

        Returns:
        -------
        Stochastic prediction of the output y
        """
        # TODO slows down a lot, maybe FeatureEngineering should return training data still as list
        x = np.array([x for x in X]).T
        x = self.x_scaler.transform(x)
        x = [np.array(i) for i in x.T.tolist()]

        # TODO unlike ANNs, GPs should provide API for vectorised .predict() and other methods
        y, std, _ = self.feat_eng._predict(x, feed_forward=lambda t: self.model.predict(t))

        y = self.y_scaler.inverse_transform(y)

        self.y_scaler.with_mean = False
        std = self.y_scaler.inverse_transform(std * np.ones(y.shape))
        self.y_scaler.with_mean = True

        return y, std

    def save_state(self, state=None, **kwargs):
        """
        Save the state of GP surrogate as a pickle file
        """
        state = self.__dict__
        save_state = super().save_state(state=state, name=self.name)

    def load_state(self):
        """
        Load the state of the GP surrogate from file
        """
        super.load_state(name=self.name)

    ##########################
    # END COMMON SUBROUTINES #
    ##########################

    def set_data_stats(self):
        """
        If the data were standardized, this stores the mean and
        standard derivation of the features in the GP_Surrogate object.
        """
        if hasattr(self.model, 'X_mean'):
            self.feat_mean = self.model.X_mean
            self.seat_Std = self.model.X_std
        else:
            self.feat_mean = 0.0 * np.ones((1, self.n_in))
            self.feat_std = 1.0 * np.ones((1, self.n_in))

        if hasattr(self.model, 'y_mean'):
            self.output_mean = self.model.y_mean
            self.output_std = self.model.y_std
        else:
            self.output_mean = 0.0 * np.ones((1, self.n_out))
            self.output_std = 1.0 * np.ones((1, self.n_out))

    def train_sequentially(self, feats=None, target=None,
                           n_iter=0, **kwargs):
        """
        Update GP surrogate with a sequeantial design scheme

        Parameters
        ----------
        feats: list of feature arrays
        target: array of target data
        n_iter: integer, number of iterations of sequential optimisation
        """

        self.set_data_stats()

        if 'acquisition_function' in kwargs:
            acq_func_arg = kwargs['acquisition_function']
            if acq_func_arg == 'poi':
                acq_func_obj = self.poi_acquisition_function
            elif acq_func_arg == 'mu':
                acq_func_obj = self.maxunc_acquisition_function
            else:
                raise NotImplementedError(
                    'This rule for sequential optimisation is not implemented, using default.')
        else:
            acq_func_obj = self.maxunc_acquisition_function

        if 'save_history' in kwargs:
            save_history = kwargs['save_history']
        else:
            save_history = False

        if save_history:
            self.design_history = []

        if self.backend == 'scikit-learn':

            """
            0) iterate for n_iter
                1) state on step n: object has X_train, X_test, their indices, model instance
                2) find set of candidates at minima of acq function X_cand; now object has
                X_train:=X_train U X_cand, X_test = X_test U_ X_cand, global inidces and set sizes updated
                3) model instance is updated : first approach to train new model for new X_train
            """

            for i in range(n_iter):

                X_new, x_new_ind_test, x_new_ind_glob = self.feat_eng.\
                    chose_feature_from_acquisition(acq_func_obj, self.X_test)
                X_new = X_new.reshape(1, -1)

                # x_new_inds = feats.index(X_new)  # feats is list of features, for this
                # has to be list of samples
                y_new = self.y_test[x_new_ind_test].reshape(1, -1)

                self.feat_eng.train_indices = np.concatenate([self.feat_eng.train_indices,
                                                              np.array(x_new_ind_glob).reshape(-1)])
                self.feat_eng.test_indices = np.delete(
                    self.feat_eng.test_indices, x_new_ind_test, 0)
                self.feat_eng.n_train += 1
                self.feat_eng.n_test -= 1

                X_train = np.concatenate([self.X_train, X_new])
                y_train = np.concatenate([self.y_train, y_new])
                X_test = np.delete(self.X_test, x_new_ind_test, 0)
                y_test = np.delete(self.y_test, x_new_ind_test, 0)

                if save_history:
                    self.design_history.append(x_new_ind_test)

                # TODO update the scaler - has to transform back to original values, then refit
                #X_train = self.x_scaler.fit_transform(X_train)
                #y_train = self.y_scaler.fit_transform(y_train)
                #X_test = self.x_scaler.transform(X_test)
                #y_test = self.y_scaler.transform(y_test)

                self.X_train = X_train
                self.y_train = y_train
                self.X_test = X_test
                self.y_test = y_test

                # self.model = es.methods.GP(X_train, y_train,
                # kernel=self.base_kernel, bias=False, noize=self.noize,
                # backend=self.backend)

                self.model.train(X_train, y_train)

        elif self.backend == 'mogp':
            pass
        else:
            raise NotImplementedError('Currently supporting only scikit-learn and mogp backend')

    def derivative_x(self, X):
        """
        Make a prediction of the derivative of output y by input x

        Args:
            X: list of feature arrays

        Returns:
        -------
        dy/dx
        """
        # TODO ideally should be a jacobian (dy_i/dx_j)_ij
        if self.backend == 'mogp':

            x = np.array([x for x in X]).T
            x = self.x_scaler.transform(x)
            x = [np.array(i) for i in x.T.tolist()]

            _, _, der = self.feat_eng._predict(x, feed_forward=lambda t: self.model.predict(t))
        else:
            raise NotImplementedError(
                "Gaussian Process derivatives w.r.t. inputs are implemented only for MOGP")

        self.y_scaler.with_mean = False
        der = self.y_scaler.inverse_transform(der)
        self.y_scaler.with_mean = True

        self.x_scaler.with_mean = False
        der = self.x_scaler.transform(der)
        self.x_scaler.with_mean = True

        return der

    def maxunc_acquisition_function(self, sample, candidates=None):
        """
        Returns the uncertainty of the model as (a posterior variance on Y) for a given sample
        Args:
            sample: a single sample from a feature array
            candidates: list of input parameter files to chose optimum from
        Returns:
            the value of uncertainty (variance) of the model
        """

        if sample.ndim == 1:
            sample = sample[None, :]

        _, uncertatinty, _ = self.model.predict(sample)

        return -1. * uncertatinty

    def poi_acquisition_function(self, sample, candidates=None):
        """
        Returns the probability of improvement for a given sample
        Args:
            sample: a single sample from a feature array
            candidates: list of input parameter files to chose optimum from
        Returns:
            the probability of improvement if a given sample will be added to the model
        """

        jitter = 1e-9
        f_star = self.output_mean

        if sample.ndim == 1:
            sample = sample[None, :]

        mu, std, d = self.model.predict(sample)
        poi = np.linalg.norm(np.divide(abs(mu - f_star), std + jitter), ord=2)

        return -poi



"""
===============================================================================
CLASS FOR THE BINNING SURROGATE PROCEDURE
-------------------------------------------------------------------------------
Reference:
N. Verheul, D. Crommelin,
"Data-driven stochastic representations of unresolved features in multiscale models"
Communications in Mathematical Sciences, 14, 5, 2016.
Code: W. Edeling
===============================================================================
"""

import numpy as np
from scipy import stats
from itertools import chain, product, cycle
import matplotlib.pyplot as plt
import sys
from scipy.spatial import ConvexHull


class Resampler:

    def __init__(self, c, r_ip1, N, N_bins, lags, store_frame_rate=1,
                 min_count=0):

        # total number of conditional variables including lagged terms
        self.N_c = c.shape[1]

        # number of unique conditioning variables (not including lags)
        self.N_covar = len(lags)

        self.N = N
        self.r_ip1 = r_ip1
        self.c = c
        self.N_bins = N_bins
        self.lags = lags
        self.max_lag = np.max(list(chain(*lags)))
        self.covar = {}

        for i in range(self.N_covar):
            self.covar[i] = []

        bins = self.get_bins(N_bins)

        # count = numver of r_ip1 samples in each bin
        # binedges = same as bins
        # binnumber = bin indices of the r_ip1 samples. A 1D array, no matter N_c
        count, binedges, binnumber = stats.binned_statistic_dd(
            c, r_ip1, statistic='count', bins=bins)

        # the unique set of binnumers which have at least one r_ip1 sample
        unique_binnumbers = np.unique(binnumber)

        # some scalars
        binnumber_max = np.max(unique_binnumbers)

        # array containing r_ip1 indices sorted PER BIN
        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        # SHOULD BE A 1D ARRAY, AN ARRAY OF SIZE [MAX(BINNUMBER), MAX(COUNT)]
        # WILL STORE MOSTLY ZEROS IN HIGHER DIMENSIONS, LEADING TO MEMORY FAILURES
        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        idx_of_bin = []

        # number of samples in each bin
        size_of_bin = []

        for b in unique_binnumbers:
            # r_ip1 indices of bin b
            tmp = np.where(binnumber == b)[0]

            idx_of_bin.append(tmp)
            size_of_bin.append(tmp.size)

        # make a 1d array of the nested arrays
        idx_of_bin = np.array(list(chain(*idx_of_bin)))
        size_of_bin = np.array(size_of_bin)

        # the starting offset of each bin in the 1D array idx_of_bin
        # offset[unique_binnumbers] will give the starting index in idex_of_bin for
        # each bin. Some entries are zero, which correspond to empty bins (except for
        # the lowest non-empty binnumber)
        offset = np.zeros(unique_binnumbers.size).astype('int')
        offset[1:] = np.cumsum(size_of_bin[0:-1])
        tmp = np.zeros(binnumber_max + 1).astype('int')
        tmp[unique_binnumbers] = offset
        self.offset = tmp

        ####################
        #empty bin handling#
        ####################

        # find indices of empty bins
        # Note: if 'full' is defined as 1 or more sample, binnumbers_nonempty is
        # the same as binnumbers_unique
        x_idx_nonempty = np.where(count > min_count)
        x_idx_nonempty_p1 = [x_idx_nonempty[i] + 1 for i in range(self.N_c)]
        binnumbers_nonempty = np.ravel_multi_index(x_idx_nonempty_p1, [len(b) + 1 for b in bins])
        N_nonempty = binnumbers_nonempty.size

        # mid points of the bins
        x_mid = [0.5 * (bins[i][1:] + bins[i][0:-1]) for i in range(self.N_c)]

        # mid points of the non-empty bins
        midpoints = np.zeros([N_nonempty, self.N_c])
        for i in range(self.N_c):
            midpoints[:, i] = x_mid[i][x_idx_nonempty[i]]

        self.bins = bins
        self.count = count
        self.binnumber = binnumber
        self.binnumbers_nonempty = binnumbers_nonempty
        self.midpoints = midpoints
        self.idx_of_bin = idx_of_bin
        self.unique_binnumbers = unique_binnumbers

        print('Connecting all possible empty bins to nearest non-empty bin...')
        self.fill_in_blanks()
        print('done.')

        # mean r per cell
        self.rmean, _, _ = stats.binned_statistic_dd(c, r_ip1, statistic='mean', bins=bins)

    # check which c_i fall within empty bins and correct binnumbers_i by
    # projecting to the nearest non-empty bin
    def check_outliers(self, binnumbers_i, c_i):

        # find out how many BINS with outliers there are
        unique_binnumbers_i = np.unique(binnumbers_i)
        idx = np.where(np.in1d(unique_binnumbers_i, self.binnumbers_nonempty) == False)[0]
        N_outlier_bins = idx.size

        if N_outlier_bins > 0:

            # index of outlier SAMPLES in binnumbers_i
            outliers_idx = np.in1d(binnumbers_i, unique_binnumbers_i[idx]).nonzero()[0]
            N_outliers = outliers_idx.size

#            if self.verbose == True:
#                print(N_outlier_bins, ' bins with', N_outliers ,'outlier samples found')

            # x location of outliers
            x_outliers = np.copy(c_i[outliers_idx])

            # find non-empty bin closest to the outliers
            closest_idx = np.zeros(N_outliers).astype('int')
            for i in range(N_outliers):
                if self.N_c == 1:
                    dist = np.linalg.norm(self.midpoints - x_outliers[i], 2, 1)
                else:
                    dist = np.linalg.norm(self.midpoints - x_outliers[i, :], 2, 1)
                closest_idx[i] = np.argmin(dist)

            binnumbers_closest = self.binnumbers_nonempty[closest_idx]

            check = self.binnumbers_nonempty[closest_idx]
            x_idx_check = np.unravel_index(check, [len(b) + 1 for b in self.bins])
            x_idx_check = [x_idx_check[i] - 1 for i in range(self.N_c)]

            # overwrite outliers in binnumbers_i with nearest non-empty binnumber
            binnumbers_closest = self.binnumbers_nonempty[closest_idx]

#            if self.verbose == True:
#                print('Moving', binnumbers_i[outliers_idx], '-->', binnumbers_closest)

            binnumbers_i[outliers_idx] = binnumbers_closest

    # create an apriori mapping between every possible bin and the nearest
    # non-empty bin. Non-empty bins will link to themselves.
    def fill_in_blanks(self):

        bins_padded = []

        # mid points of all possible bins, including ghost bins
        for i in range(self.N_c):
            dx1 = self.bins[i][1] - self.bins[i][0]
            dx2 = self.bins[i][-1] - self.bins[i][-2]

            # pad the beginning and end of current 1D bin with extrapolated values
            bin_pad = np.pad(self.bins[i], (1, 1), 'constant', constant_values=(
                self.bins[i][0] - dx1, self.bins[i][-1] + dx2))
            bins_padded.append(bin_pad)

        # compute the midpoints of the padded bins
        x_mid_pad = [0.5 * (bins_padded[i][1:] + bins_padded[i][0:-1]) for i in range(self.N_c)]
        self.x_mid_pad_tensor = np.array(list(product(*x_mid_pad)))

        # total number bins
        self.max_binnumber = self.x_mid_pad_tensor.shape[0]

        mapping = np.zeros(self.max_binnumber).astype('int')

        for i in range(self.max_binnumber):

            if np.mod(i, 10) == 0:
                progress(i, self.max_binnumber)

            # bin is nonempty, just use current idx
            if np.in1d(i, self.unique_binnumbers):
                mapping[i] = i
            # bin is empty, find nearest non-empty bin
            else:
                binnumbers_i = np.array([i])
                self.check_outliers(binnumbers_i, self.x_mid_pad_tensor[i].reshape([1, self.N_c]))
                mapping[i] = binnumbers_i[0]

        self.mapping = mapping

    # visual representation of a 2D binning object. Also shows the mapping
    # between empty to nearest non-empty bins.
    def plot_2D_binning_object(self):

        if self.N_c != 2:
            print('Only works for N_c = 2')
            return

        fig = plt.figure()
        ax = fig.add_subplot(111, xlabel=r'conditioning variable 1',
                             ylabel=r'conditioning variable 2')

        # plot bins and (c1, c2) which corresponding to a r sample point
        ax.plot(self.c[:, 0], self.c[:, 1], '+', color='lightgray', alpha=0.3)
        ax.vlines(self.bins[0], np.min(self.c[:, 1]), np.max(self.c[:, 1]))
        ax.hlines(self.bins[1], np.min(self.c[:, 0]), np.max(self.c[:, 0]))

        ax.plot(self.x_mid_pad_tensor[:, 0], self.x_mid_pad_tensor[:, 1], 'g+')

        # plot the mapping
        for i in range(self.max_binnumber):
            ax.plot([self.x_mid_pad_tensor[i][0], self.x_mid_pad_tensor[self.mapping[i]][0]], [
                    self.x_mid_pad_tensor[i][1], self.x_mid_pad_tensor[self.mapping[i]][1]], 'b', alpha=0.6)

        ax.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))
        ax.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
        plt.tight_layout()
        plt.show()

    def plot_2D_shadow_manifold(self, X):

        if self.N_c != 2:
            print('Only works for N_c = 2')
            return

        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.plot(X[:, 0], X[:, 1], '+', color='lightgray', alpha=0.3)

        colors = cycle(['--r', '--g', '--b', '--m', '--k'])

        for idx in self.unique_binnumbers:

            idx_i = np.where(self.binnumber == idx)[0]

            if idx_i.size >= 3:
                marker = next(colors)
                points = X[idx_i, :]
                hull = ConvexHull(points)
                ax.plot(points[hull.vertices, 0], points[hull.vertices, 1], marker)
                ax.plot([points[hull.vertices[0], 0], points[hull.vertices[-1], 0]],
                        [points[hull.vertices[0], 1], points[hull.vertices[-1], 1]], marker)

                x_mid = np.mean(points[hull.vertices, :], axis=0)
                ax.text(x_mid[0], x_mid[1], str(idx))
            # ax.plot(X[idx_i, 0], X[idx_i, 1], '+')
            else:
                x_mid = np.mean(X[idx_i, :], axis=0)
                ax.text(x_mid[0], x_mid[1], str(idx))

    # the data-driven model for the unresolved scales
    # Given c_i return r at time i+1 (r_ip1)
    def get_sample(self, c_i, n_mc=1):

        # find in which bins the c_i samples fall
        _, _, binnumbers_i = stats.binned_statistic_dd(c_i, np.zeros(self.N), bins=self.bins)

        # dynamically corrects binnumbers_i if outliers are found
        #self.check_outliers(binnumbers_i, c_i)

        # static correction for outliers, using precomputed mapping array
        binnumbers_i = self.mapping[binnumbers_i]

        #self.check_outliers(binnumbers_i, c_i)

        # convert 1D binnumbers_i to equivalent ND indices
        x_idx = np.unravel_index(binnumbers_i, [len(b) + 1 for b in self.bins])
        x_idx = [x_idx[i] - 1 for i in range(self.N_c)]
        x_idx = tuple(x_idx)
        
        # random integers between 0 and max bin count for each index in binnumbers_i
        I = np.floor(self.count[x_idx].reshape([self.N, 1]) *
                     np.random.rand(self.N, n_mc)).astype('int')

        # the correct offset for the 1D array idx_of_bin
        start = self.offset[binnumbers_i]

        # get random r sample from each bin indexed by binnumbers_i
        r = np.zeros([n_mc, self.N, self.N])

        for i in range(n_mc):
            r[i, :, :] = self.r_ip1[self.idx_of_bin[start + I[:, i]]].reshape([self.N, self.N])

        return np.mean(r, 0)  # , binnumbers_i

    # the data-driven model for the unresolved scales
    # Given c_i return bin averaged r at time i+1
    def get_mean_sample(self, c_i):

        # find in which bins the c_i samples fall
        _, _, binnumbers_i = stats.binned_statistic_dd(c_i, np.zeros(self.N), bins=self.bins)

        # dynamically corrects binnumbers_i if outliers are found
        #self.check_outliers(binnumbers_i, c_i)

        # static correction for outliers, using precomputed mapping array
        binnumbers_i = self.mapping[binnumbers_i]

        # convert 1D binnumbers_i to equivalent ND indices
        x_idx = np.unravel_index(binnumbers_i, [len(b) + 1 for b in self.bins])
        x_idx = [x_idx[i] - 1 for i in range(self.N_c)]

        # , self.rstd[x_idx].reshape([self.N, self.N])
        return self.rmean[x_idx].reshape([self.N, self.N])

    # append the covariates supplied to the binning object during simulation
    # to self.covars
    # Note: use list to dynamically append, array is very slow
    def append_covar(self, c_i):

        for i in range(self.N_covar):
            self.covar[i].append(c_i[:, i])

            # if max number of covariates is reached, remove first item
            if len(self.covar[i]) > self.max_lag:
                self.covar[i].pop(0)

    # return lagged covariates, assumes constant lag
    # Note: typecast to array if spatially varying lag is required
    def get_covar(self):

        c_i = np.zeros([self.N, self.N_c])

        idx = 0

        for i in range(self.N_covar):
            for lag in self.lags[i]:
                c_i[:, idx] = self.covar[i][-lag]
                idx += 1

        return c_i

    def print_bin_info(self):
        print('-------------------------------')
        print('Total number of samples= ', self.r_ip1.size)
        # print('Total number of bins = ', self.N_bins**self.N_c)
        print('Total number of non-empty bins = ', self.binnumbers_nonempty.size)
        # print('Percentage filled = ', np.double(self.binnumbers_nonempty.size)/self.N_bins**self.N_c*100., ' %')
        print('-------------------------------')

    # compute the uniform bins of the conditional variables in c
    def get_bins(self, N_bins):

        bins = []

        for i in range(self.N_c):
            bins.append(np.linspace(np.min(self.c[:, i]), np.max(self.c[:, i]), N_bins[i] + 1))

        return bins


def progress(count, total, status=''):
    """
    progress bar for the command line
    Source: https://gist.github.com/vladignatyev/06860ec2040cb497f0f3
    """
    bar_len = 60
    filled_len = int(round(bar_len * count / float(total)))

    percents = round(100.0 * count / float(total), 1)
    bar = '=' * filled_len + '-' * (bar_len - filled_len)

    sys.stdout.write('[%s] %s%s ...%s\r' % (bar, percents, '%', status))
    sys.stdout.flush()  # As suggested by Rom Ruben (see: http://stackoverflow.com/questions/3173320/text-progress-bar-in-the-console/27871113#comment50529068_27871113)



import numpy as np
from itertools import product
from scipy.optimize import minimize


class GaussianProcess():

    def __init__(self):
        self.n_x_dim = 1

    def set_kernel(self, kernel='sq_exp'):
        if kernel == 'gibbs':
            self.kernel = gibbs_ns_kernel
        else:
            self.kernel = sq_exp_kernel_function

        self.sigma_f = 1.0
        self.sigma_n = 0.1
        self.l = 1.0

    def set_covariance(self, K):
        self.K = K
        self.n = K.shape[0]

    def fit_cov(self, X, covariance='regular'):

        K = [self.kernel(i, j, sigma_f=self.sigma_f, l=self.l) for (i, j) in product(X, X)]
        K = np.array(K).reshape(self.n, self.n)
        self.set_covariance(K)

        K_inv_tot = np.linalg.inv(self.K + (self.sigma_n ** 2) * np.eye(self.n))
        self.K_inv_tot = K_inv_tot

    def optmize_hyperparameters(self, X_train, y_train):
        """
        Optimizes hyperparamter values for minimum of R^2 score on training dataset
        #TODO has to optimise form MLE or MAP
        Uses scipy .minimize() to find the optimum
        Reassigns the attributes of the object after optimization
        """
        self.X_train = X_train
        self.y_train = y_train

        hp_curval = np.array([self.sigma_f, self.sigma_n, self.l])

        hp_optval_res = minimize(self.r2_score_hp, hp_curval, options={'maxiter': 100})

        hp_optval = hp_optval_res.x

        self.sigma_f = hp_optval[0]
        self.sigma_n = hp_optval[1]
        self.l = hp_optval[2]

    def fit(self, X, y):

        self.n = X.shape[0]
        self.X = X
        self.y = y

        self.fit_cov(X)

        self.optmize_hyperparameters(X, y)

    def predict_mean(self, X_new):

        n_star = X_new.shape[0]
        K_star = [
            self.kernel(
                i,
                j,
                sigma_f=self.sigma_f,
                l=self.l) for (
                i,
                j) in product(
                X_new,
                self.X)]
        K_star = np.array(K_star).reshape(n_star, self.n)

        f_bar_star = np.dot(K_star, np.dot(self.K_inv_tot, self.y.reshape(self.n, self.n_x_dim)))
        return f_bar_star

    def predict_var(self, X_new):

        n_star = X_new.shape[0]

        K_star2 = [
            self.kernel(
                i,
                j,
                sigma_f=self.sigma_f,
                l=self.l) for (
                i,
                j) in product(
                X_new,
                X_new)]
        K_star2 = np.array(K_star2).reshape(n_star, n_star)

        K_star = [
            self.kernel(
                i,
                j,
                sigma_f=self.sigma_f,
                l=self.l) for (
                i,
                j) in product(
                X_new,
                self.X)]
        K_star = np.array(K_star).reshape(n_star, self.n)

        cov_f_star = K_star2 - np.dot(K_star, np.dot(self.K_inv_tot, K_star.T))
        var_f_star = np.diag(cov_f_star)

        return var_f_star

    def predict(self, X, return_std=True):

        y_mean = self.predict_mean(X)

        if return_std:
            y_var = self.predict_var(X)
            return y_mean, y_var

        return y_mean

    def r2_score(self, X, y):
        f = self.predict_mean(X)
        y_mean = y.mean()
        r2 = 1 - np.multiply(y - f, y - f).sum() / (np.multiply(y - y_mean, y - y_mean).sum())
        return r2

    def r2_score_hp(self, hpval):
        self.sigma_f = hpval[0]
        self.sigma_n = hpval[1]
        self.l = hpval[2]
        self.fit_cov(self.X_train)
        return self.r2_score(self.X_train, self.y_train)


def gibbs_ns_kernel(x, y, l, l_func=lambda x: x):
    pref_val = 1.0
    exp_arg = 0.0
    for d in x.shape[1]:
        # multiplicatve prefactor
        pref_val *= np.sqrt((2 * l_func(x) * l_func(y)) / (l_func(x)**2 + l_func(y)**2))
        # exponential argument
        exp_arg += (x - y)**2 / (l_func(x)**2 + l_func(y)**2)
    exp_val = np.exp(-exp_arg)
    return pref_val * exp_val


def sq_exp_kernel_function(x, y, sigma_f=1., l=1.):
    """Define squared exponential kernel function."""
    kernel = sigma_f * np.exp(- (np.linalg.norm(x - y)**2) / (2 * l**2))
    return kernel



"""
CLASS TO PERFORM ANALYSIS ON RESULTS FROM AN ARTIFICIAL NEURAL NETWORK.
"""
import numpy as np
from .base import BaseAnalysis


class ANN_analysis(BaseAnalysis):
    """
    ANN analysis class
    """

    def __init__(self, ann_surrogate):
        print('Creating ANN_analysis object')
        self.ann_surrogate = ann_surrogate

    def sensitivity_measures(self, feats, norm=True):
        """
        Compute global derivative-based sensitivity measures using the
        derivative of squared L2 norm of the output, computing usoing back propagation.
        Integration of the derivatives over the input space is done via MC on the provided
        input features in feats.

        Parameters
        ----------
        feats : array
            An array of input parameter values.

        norm : Boolean, optional, default is True.
            Compute the gradient of ||y||^2_2. If False it computes the gradient of
            y, if y is a scalar. If False and y is a vector, the resulting gradient is the
            column sum of the full Jacobian matrix.

        Returns
        -------
        idx : array
            Indices corresponding to input variables, ordered from most to least
            influential.

        """

        # standardize the features
        feats = (feats - self.ann_surrogate.feat_mean) / self.ann_surrogate.feat_std
        N = feats.shape[0]
        # Set the batch size to 1
        self.ann_surrogate.neural_net.set_batch_size(1)
        # initialize the derivatives
        self.ann_surrogate.neural_net.d_norm_y_dX(feats[0].reshape([1, -1]),
                                                  norm=norm)
        # compute the squared gradient
        norm_y_grad_x2 = self.ann_surrogate.neural_net.layers[0].delta_hy**2
        # compute the mean gradient
        mean = norm_y_grad_x2 / N
        # loop over all samples
        for i in range(1, N):
            # compute the next (squared) gradient
            self.ann_surrogate.neural_net.d_norm_y_dX(feats[i].reshape([1, -1]))
            norm_y_grad_x2 = self.ann_surrogate.neural_net.layers[0].delta_hy**2
            mean += norm_y_grad_x2 / N
        # order parameters from most to least influential based on the mean
        # squared gradient
        idx = np.fliplr(np.argsort(np.abs(mean).T))
        print('Parameters ordered from most to least important:')
        print(idx)
        return idx, mean

    def get_errors(self, feats, data, relative=True, return_predictions=False):
        """
        Get the training and test error of the ANN surrogate to screen. This method
        uses the ANN_Surrogate.get_dimensions() dictionary to determine where the split
        between training and test data is:

            [0,1,...,n_train,n_train+1,...,n_samples]

        Hence the last entries are used as test data, and feats and data should structured as
        such.

        Parameters
        ----------
        feats : array, size = [n_samples, n_feats]
            The features.
        data : array, size = [n_samples, n_out]
            The data.
        relative : boolean, default is True
            Compute relative instead of absolute errors.
        return_predictions : boolean, default is False
            Also return the train and test predictions.

        Returns
        -------
        err_train, err_test : float
            The training and test errors

        """
        dims = self.ann_surrogate.get_dimensions()
        # run the trained model forward at training locations
        n_mc = dims['n_train']
        train_pred = np.zeros([n_mc, dims['n_out']])
        for i in range(n_mc):
            train_pred[i, :] = self.ann_surrogate.predict(feats[i])

        train_data = data[0:dims['n_train']]
        if relative:
            err_train = np.mean(np.linalg.norm(train_data - train_pred, axis=0) /
                                np.linalg.norm(train_data, axis=0), axis=0)
            print("Relative training error = %.4f %%" % (err_train * 100))
        else:
            err_train = np.mean(np.linalg.norm(train_data - train_pred, axis=0), axis=0)
            print("Training error = %.4f " % (err_train))

        # run the trained model forward at test locations
        test_pred = np.zeros([dims['n_test'], dims['n_out']])
        for idx, i in enumerate(range(dims['n_train'], dims['n_samples'])):
            test_pred[idx] = self.ann_surrogate.predict(feats[i])
        test_data = data[dims['n_train']:]
        if relative:
            err_test = np.mean(np.linalg.norm(test_data - test_pred, axis=0) /
                               np.linalg.norm(test_data, axis=0), axis=0)
            print("Relative test error = %.4f %%" % (err_test * 100))
        else:
            err_test = np.mean(np.linalg.norm(test_data - test_pred, axis=0), axis=0)
            print("Test error = %.4f " % (err_test))

        if not return_predictions:
            return err_train, err_test
        else:
            return err_train, err_test, train_pred, test_pred



"""
CLASS TO PERFORM ANALYSIS ON RESULTS FROM A QUANTIZED SOFTMAX SURROGATE.
"""
from .base import BaseAnalysis
from scipy import stats
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import animation


class QSN_analysis(BaseAnalysis):
    """
    QSN analysis class
    """

    def __init__(self, qsn_surrogate, **kwargs):
        print('Creating QSN_analysis object')
        self.qsn_surrogate = qsn_surrogate

    def get_classification_error(self, **kwargs):
        """
        Compute the misclassification error of the QSN surrogate.

        Parameters
        ----------


        Returns
        -------
        Prints classification error to screen for every softmax layer

        """

        if 'X' not in kwargs:
            X = self.qsn_surrogate.neural_net.X
            y = self.qsn_surrogate.neural_net.y
        else:
            X = kwargs['X']
            y = kwargs['y']

        self.qsn_surrogate.neural_net.compute_misclass_softmax(X=X, y=y)

    # def __bin_data(self, data):
    def bin_data(self, data):

        n_bins = self.qsn_surrogate.n_bins
        n_vars = self.qsn_surrogate.feat_eng.n_vars

        n_samples = data.shape[0]

        binnumbers = np.zeros([n_samples, n_vars]).astype('int')
        y_idx = np.zeros([n_samples, n_bins * n_vars])

        for i in range(n_vars):

            bins = self.qsn_surrogate.feat_eng.bins[i]

            _, _, binnumbers[:, i] = \
                stats.binned_statistic(data[:, i], np.zeros(n_samples),
                                       statistic='count', bins=bins)

            unique_binnumber = np.unique(binnumbers[:, i])

            offset = i * n_bins

            for j in unique_binnumber:
                idx = np.where(binnumbers[:, i] == j)
                y_idx[idx, offset + j - 1] = 1.0

        return y_idx

    def get_errors(self, feats, data, relative=True, return_predictions=False):
        """
        Get the training and test error of the ANN surrogate to screen. This method
        uses the ANN_Surrogate.get_dimensions() dictionary to determine where the split
        between training and test data is:

            [0,1,...,n_train,n_train+1,...,n_samples]

        Hence the last entries are used as test data, and feats and data should structured as
        such.

        Parameters
        ----------
        feats : array, size = [n_samples, n_feats]
            The features.
        data : array, size = [n_samples, n_out]
            The data.
        relative : boolean, default is True
            Compute relative instead of absolute errors.
        return_predictions : boolean, default is False
            Also return the train and test predictions.

        Returns
        -------
        err_train, err_test : float
            The training and test errors

        """
        dims = self.qsn_surrogate.get_dimensions()

        # y_idx_train, y_idx_test = self.__bin_data(data)
        y_idx_train = self.bin_data(data[0:dims['n_train']])
        y_idx_test = self.bin_data(data[dims['n_train']:])

        X_train = (feats[0:dims['n_train'], :] - self.qsn_surrogate.feat_mean) / \
            self.qsn_surrogate.feat_std
        X_test = (feats[dims['n_train']:, :] - self.qsn_surrogate.feat_mean) / \
            self.qsn_surrogate.feat_std

        self.qsn_surrogate.neural_net.compute_misclass_softmax(X=X_train, y=y_idx_train)
        self.qsn_surrogate.neural_net.compute_misclass_softmax(X=X_test, y=y_idx_test)

    def get_KL_errors(self, feats, ref_distributions):

        n_samples = feats.shape[0]
        n_softmax = self.qsn_surrogate.n_softmax

        for i in range(n_samples):
            o_i, _, _ = self.qsn_surrogate.neural_net.get_softmax(feats[i].reshape([1, -1]))

            KL_div = np.zeros(n_softmax)
            for j, y_j in enumerate(np.split(ref_distributions[i], n_softmax)):

                idx_gt0 = np.where(y_j > 0.0)[0]
                KL_div[j] = -np.sum(y_j[idx_gt0] * np.log(o_i[j][idx_gt0] / y_j[idx_gt0]))

            print("KL divergence sample %d = %s" % (i, KL_div))

    # def make_movie(self, n_frames=500):
    #     """
    #     Makes a move using the training data. Left subplot shows the evolution of the
    #     kernel-density estimate, and the right subplot show the time series of the data
    #     and the random samples drawm from the kernel-density estimate. Saves the movie
    #     to a .gif file.

    #     Parameters

    #     n_frames (int): default is 500
    #         The number of frames to use in the movie.

    #     Returns: None
    #     """

    #     # get the (normalized, time-lagged) training data from the neural network
    #     X = self.qsn_surrogate.neural_net.X
    #     y = self.qsn_surrogate.neural_net.y

    #     print('===============================')
    #     print('Making movie...')

    #     # list to store the movie frames in
    #     ims = []
    #     fig = plt.figure(figsize=[8, 4])
    #     ax1 = fig.add_subplot(121, xlabel=r'$B_k$', ylabel=r'', yticks=[])
    #     ax2 = fig.add_subplot(122, xlabel=r'time', ylabel=r'$B_k$')
    #     plt.tight_layout()

    #     # number of features
    #     n_feat = X.shape[1]
    #     # number of softmax layers
    #     n_softmax = self.qsn_surrogate.n_softmax

    #     # allocate memory
    #     samples = np.zeros([n_frames, n_softmax])

    #     # make movie by evaluating the network at TRAINING inputs
    #     for i in range(n_frames):

    #         # draw a random sample from the network
    #         o_i, idx_max, _ = self.qsn_surrogate.neural_net.get_softmax(X[i].reshape([1, n_feat]))
    #         self.qsn_surrogate.sampler.resample(idx_max)
    #         if np.mod(i, 100) == 0:
    #             print('i =', i, 'of', n_frames)

    #         # create a single frame, store in 'ims'
    #         plt2 = ax1.plot(range(self.qsn_surrogate.n_bins),
    #                         np.zeros(self.qsn_surrogate.n_bins),
    #                         o_i[0], 'b', label=r'conditional pmf')
    #         plt3 = ax1.plot(y[i][0], 0.0, 'ro', label=r'data')
    #         plt4 = ax2.plot(y[0:i, 0], 'ro', label=r'data')
    #         plt5 = ax2.plot(samples[0:i, 0], 'g', label='random sample')

    #         if i == 0:
    #             ax1.legend(loc=1, fontsize=9)
    #             ax2.legend(loc=1, fontsize=9)

    #         ims.append((plt2[0], plt3[0], plt4[0], plt5[0],))

    #     # make a movie of all frame in 'ims'
    #     im_ani = animation.ArtistAnimation(fig, ims, interval=20,
    #                                        repeat_delay=2000, blit=True)
    #     im_ani.save('kvm.gif')
    #     print('done. Saved movie to qsn.gif')



"""
CLASS TO PERFORM ANALYSIS ON RESULTS FROM A GAUSSIAN PROCESS SURROGATE
"""

from .base import BaseAnalysis
import matplotlib.pyplot as plt
import numpy as np
from scipy import stats
from sklearn.metrics import mean_squared_error as mse
from mpl_toolkits.axes_grid1 import make_axes_locatable


class GP_analysis(BaseAnalysis):
    """
    GP analysis class
    """

    def __init__(self, gp_surrogate, **kwargs):

        print('Creating GP_analysis object')

        self.gp_surrogate = gp_surrogate

    def plot_err(self, error, name, original=None):
        plt.ioff()
        plt.title(name)
        plt.plot(range(len(error)), error, '.', label='GP metamodel error', color='b')
        if original is not None:
            plt.plot(original, '.', label='Simulation model', color='green')
        plt.xlabel('Run number')
        plt.ylabel('Results of {}'.format(name))
        plt.legend()
        plt.grid('major')
        plt.yscale("symlog")
        plt.savefig('gp_abs_err.png')
        plt.close()

    def plot_res(self,
                 x_test, y_test_pred, y_test_orig,
                 x_train, y_train_pred, y_train_orig,
                 y_var_pred_test=None, y_var_pred_train=None,
                 name='', num='1', type_train='rand', train_n=10, out_color='b', output_folder=''):

        plt.ioff()
        plt.figure(figsize=(8, 11))

        y_pred = np.zeros(y_test_pred.shape[0] + y_train_pred.shape[0])
        y_pred[x_train] = y_train_pred
        y_pred[x_test] = y_test_pred

        y_orig = np.zeros(y_test_orig.shape[0] + y_train_orig.shape[0])
        y_orig[x_train] = y_train_orig
        y_orig[x_test] = y_test_orig

        # --- Plotting prediction vs original test data
        plt.subplot(311)
        plt.title('{} training sample of size: {}'.format(type_train, str(train_n)))

        plt.plot(x_test, y_test_orig, '.', label='Simulation, test', color='red')
        plt.plot(x_train, y_train_orig, '*', label='Simulation, train', color='green')

        if y_var_pred_test is not None and y_var_pred_train is not None:
            plt.errorbar(
                x=x_test,
                y=y_test_pred,
                yerr=1.96 *
                y_var_pred_test,
                label='GP metamodel, test',
                fmt='+')
            plt.errorbar(
                x=x_train,
                y=y_train_pred,
                yerr=1.96 *
                y_var_pred_train,
                label='GP metamodel, train',
                fmt='+')
        else:
            plt.plot(x_test, y_test_pred, '.', label='GP metamodel', color=out_color)
            plt.plot(x_train, y_train_pred, '*', label='GP metamodel', color=out_color)

        plt.xlabel('Run number')
        plt.ylabel('Prediction of {}'.format(name))
        plt.legend()
        plt.grid()
        plt.yscale("symlog")

        # --- Plotting absolute errors
        plt.subplot(312)
        err_abs = y_pred - y_orig

        plt.plot(err_abs, '.', color=out_color)

        plt.grid()
        plt.yscale("symlog")
        plt.xlabel('Run number')
        plt.ylabel('Error')

        # print('Indices of test data where absolute error is larger than {} : {} '
        #       .format(2e5, np.where(abs(err_abs) > 2e5)[0]))

        # --- Plotting relative errors
        plt.subplot(313)
        plt.plot(np.fabs(y_pred - y_orig) / y_orig * 100, '.', color=out_color)
        plt.grid()
        plt.xlabel('Run number')
        plt.ylabel('Relative error (%)')
        plt.yscale("log")
        plt.tight_layout()
        plt.savefig(
            output_folder +
            'GP_prediction_' +
            num +
            '_' +
            type_train +
            '_' +
            str(train_n) +
            '.png',
            bbox_inches='tight',
            dpi=100)
        plt.clf()
        plt.close()

    def plot_pdfs(self, y_dom, y_pdf, y_dom_sur, y_pdf_sur,
                  y_dom_tr=None, y_pdf_tr=None, y_dom_tot=None, y_pdf_tot=None,
                  names=['GEM testing', 'GP', 'GEM training', 'GEM', ],
                  qoi_names=['TiFl'], filename='pdfs'):

        fig, ax = plt.subplots(figsize=[7, 7])

        ax.plot(y_dom, y_pdf, 'r-', label=names[0])
        ax.plot(y_dom_sur, y_pdf_sur, 'b', label=names[1])

        if y_dom_tr is not None and y_pdf_tr is not None:
            ax.plot(y_dom_tr, y_pdf_tr, 'g-', label=names[2])

        if y_dom_tot is not None and y_pdf_tot is not None:
            ax.plot(y_dom_tot, y_pdf_tot, 'k-', label=names[3])

        ax.set_xlabel(qoi_names[0])
        ax.set_ylabel('pdf')

        ax.xaxis.grid(True, which='major')
        ax.yaxis.grid(True, which='major')

        plt.legend(loc='best')
        plt.title('PDF of simulated and predicted target values')
        plt.savefig(filename + '.png')
        plt.close()

    def plot_2d_design_history(self, x_test=None, y_test=None):

        if not hasattr(self.gp_surrogate, 'design_history'):
            raise RuntimeWarning('This surrogate has no recorded history of sequential design')
            return

        x_train = self.gp_surrogate.x_scaler.inverse_transform(self.gp_surrogate.X_train)
        x1tr = x_train[:, 0]
        x2tr = x_train[:, 1]
        if x_test is not None:
            x1ts = x_test[:, 0]
            x2ts = x_test[:, 1]
        else:
            x1ts = x1tr
            x2ts = x1tr

        ytr = self.gp_surrogate.y_scaler.inverse_transform(self.gp_surrogate.y_train)
        ytr = ytr.reshape(-1)
        if y_test is not None:
            yts = y_test
            yts = yts.reshape(-1)
        else:
            yts = ytr

        fig, ax = plt.subplots()

        cntr1 = ax.tricontourf(x1ts, x2ts, yts, levels=12, cmap="RdBu_r")
        divider = make_axes_locatable(ax)
        cax1 = divider.append_axes("right", size="8%", pad=0.1)
        cbar1 = fig.colorbar(cntr1, cax1)
        ax.set_title('Ground simulation results', pad=10.0)
        ax.set_xlabel(r'$x_{1}$')
        ax.set_ylabel(r'$ x_{2}$')
        # cbar1.set_label(r'$y$')
        ax.set_aspect('equal')

        ax.plot(x1tr, x2tr, 'ko', ms=3)
        n_iter = len(self.gp_surrogate.design_history)
        for num, ind in enumerate(self.gp_surrogate.design_history):
            #p_i = (x1[ind], x2[ind])
            p_i = (x1tr[-n_iter + num], x1tr[-n_iter + num])
            ax.annotate(str(num), p_i, textcoords="offset points", xytext=(0, 10), ha='center')

        # plt.tight_layout()
        # plt.subplots_adjust()

        plt.show()
        plt.savefig('surorrogate_seq_des.png')
        plt.close()

    def get_r2_score(self, X, y):
        """
        Compute R^2 score of the regression for the test data
        Returns: vale of R^2 score
        """

        # f = gpr.predict(X)
        # y_mean = y.mean()
        # r2 = 1 - np.multiply(y - f, y - f).sum() / (np.multiply(y - y_mean, y - y_mean).sum())

        return self.gp_surrogate.model.instance.score(X, y)

    def get_regression_error(
            self,
            X_test,
            y_test,
            X_train=None,
            y_train=None,
            index=None,
            **kwargs):
        """
        Compute the regression RMSE error of GP surrogate.
        Prints RMSE regression error and plots the errors of the prediction for different simulation runs/samples
        # TODO add correlation analysis

        Args:
            index: array, indices of the dataset to form a test set

        Returns:
            None
        """

        if index is not None:
            X_test = self.gp_surrogate.model.X[index]
            y_test = self.gp_surrogate.model.y[index]

        x_test_inds = self.gp_surrogate.feat_eng.test_indices
        x_train_inds = self.gp_surrogate.feat_eng.train_indices

        print("Prediction of new QoI")
        # TODO make predict call work on a (n_samples, n_features) np array
        y_pred = [self.gp_surrogate.predict(X_test[i, :].reshape(-1, 1))[0]
                  for i in range(X_test.shape[0])]
        y_var_pred = [self.gp_surrogate.predict(X_test[i, :].reshape(-1, 1))[1]
                      for i in range(X_test.shape[0])]
        y_t_len = len(y_test)

        y_pred_train = [self.gp_surrogate.predict(
            X_train[i, :].reshape(-1, 1))[0] for i in range(X_train.shape[0])]
        y_var_pred_train = [self.gp_surrogate.predict(
            X_train[i, :].reshape(-1, 1))[1] for i in range(X_train.shape[0])]

        # reshape the resulting arrays
        y_pred = np.squeeze(np.array(y_pred), axis=1)
        y_pred_train = np.squeeze(np.array(y_pred_train), axis=1)

        y_var_pred = np.squeeze(np.array(y_var_pred), axis=1)
        y_var_pred_train = np.squeeze(np.array(y_var_pred_train), axis=1)

        # check if we a working with a vector or scalar QoI, if vector -> consider
        # only the first component
        y_test_plot = y_test
        y_train_plot = y_train
        if y_pred.shape[1] != 1:
            y_test_plot = y_test[:, [0]]
            y_train_plot = y_train[:, [0]]
            y_pred = y_pred[:, 0]
            y_pred_train = y_pred_train[:, 0]
            y_var_pred = y_var_pred[:, 0]
            y_var_pred_train = y_var_pred_train[:, 0]

        if len(y_pred.shape) == 1:
            y_pred = y_pred.reshape(-1, 1)
            y_pred_train = y_pred_train.reshape(-1, 1)

        # calculate the errors
        # test data appears smaller in length (by 1)
        err_abs = np.subtract(y_pred[:y_t_len, :], y_test)
        err_rel = np.divide(err_abs, y_test)

        print("Printing and plotting the evaluation results")
        self.plot_err(err_rel[:, 0], y_test[:, 0],
                      'rel. err. of prediction mean for test dataset in Ti fluxes')

        train_n = self.gp_surrogate.feat_eng.n_samples - y_t_len
        self.plot_res(x_test_inds, y_pred[:y_t_len, 0], y_test_plot[:, 0],
                      x_train_inds, y_pred_train[:, 0], y_train_plot[:, 0],
                      y_var_pred, y_var_pred_train,
                      r'$Y_i$', num='1', type_train='rand',
                      train_n=train_n, out_color='b')

        r2_test = self.get_r2_score(X_test, y_test)
        print('R2 score for the test data is : {:.3}'.format(r2_test))

        print('MSE of the GPR prediction is: {:.3}'.format(
            mse(y_pred[:y_t_len, 0], y_test_plot[:, 0])))

        self.y_pred = y_pred
        self.err_abs = err_abs
        self.err_rel = err_rel
        self.r2_test = r2_test



"""
Deep active subspace surrogate.
"""

import numpy as np

from .DAS_Layer import DAS_Layer
from .Layer import Layer
from .NN import ANN


class DAS_network(ANN):
    """
    Deep active subspace surrogate.

    Method:
        Tripathy, Rohit, and Ilias Bilionis. "Deep active subspaces: A scalable
        method for high-dimensional uncertainty propagation."
        ASME 2019 International Design Engineering Technical Conferences and
        Computers and Information in Engineering Conference.
        American Society of Mechanical Engineers Digital Collection, 2019.
    """

    def __init__(self, X, y, d, alpha=0.001, decay_rate=1.0, decay_step=10**5,
                 beta1=0.9, beta2=0.999, lamb=0.0, n_out=1,
                 param_specific_learn_rate=True, loss='squared',
                 activation='tanh', activation_out='linear', activation_das='linear',
                 n_softmax=0, n_layers=2, n_neurons=16,
                 bias=True, batch_size=1, batch_norm=False, 
                 save=True, name='DAS', on_gpu=False,
                 standardize_X=True, standardize_y=True, **kwargs):
        """
        Initialize the Deep Active Subspace surrogate object.

        Parameters
        ----------
        X : array
            The input features.
        y : array
            The target data.
        d : int
            The assumed dimension of the active subspace.
        alpha : float, optional
            The learning rate. The default is 0.001.
        decay_rate : float, optional
            Factor multiplying the decay rate every decay_step iterations.
            The default is 1.0.
        decay_step : int, optional
            The number of training iterations after which decay_rate is lowered.
            The default is 10**5.
        beta1 : float, optional
            Momentum parameter controlling the moving average of the loss gradient.
            Used for the parameter-specific learning rate. The default is 0.9.
        beta2 : float, optional
            Parameter controlling the moving average of the squared gradient.
            Used for the parameter-specific learning rate. The default is 0.999.
        lamb : float, optional
            L2 weight regularization parameter. The default is 0.0.
        n_out : int, optional
            The number of output neurons. The default is 1.
        param_specific_learn_rate : boolean, optional
            Use parameter-specific learing rate. The default is True.
        loss : string, optional
            The name of the loss function. The default is 'squared'.
        activation : string, optional
            The name of the activation function of the hidden layers.
            The default is 'tanh'.
        activation_out : string, optional
            The name of the activation function of the output layer.
            The default is 'linear'.
        n_softmax : int, optional
            The number of softmax layers attached to the output. The default is 0.
        n_layers : int, optional
            The number of layers, not counting the input layer. The default is 2.
        n_neurons : int, optional
            The number of neurons per hidden layer. The default is 16.
        bias : boolean, optional
            Use a bias neuron. The default is True.
        batch_size : int, optional
            The size of the mini batch. The default is 1.
        batch_norm : boolean or list of booleans, optional
            Use batch normalization in hidden layers. Not used in the DAS layer.
            The default is False.
        save : boolean, optional
            Save the neural network to a pickle file after training.
            The default is True.
        name : string, optional
            The name of the neural network. The default is 'DAS'.
        on_gpu : boolean, optional
            Train the neural network on a GPU using cupy. NOT IMPLEMENTED IN THIS VERSION.
            The default is False.
        standardize_X : boolean, optional
            Standardize the features. The default is True.
        standardize_y : boolean, optional
            Standardize the target data. The default is True.


        Returns
        -------
        None.

        """
        # the dimension of the active subspace
        self.d = d

        # the activation of the DAS layer
        self.activation_das = activation_das
        
        # set all the common parameters via the parent ANN class,
        # but overwrite the init_network subsroutine in this class
        super().__init__(X, y, alpha=alpha, decay_rate=decay_rate,
                         decay_step=decay_step, beta1=beta1, beta2=beta2,
                         lamb=lamb, n_out=n_out, loss=loss, activation=activation,
                         activation_out=activation_out, n_softmax=n_softmax,
                         n_layers=n_layers, n_neurons=n_neurons, bias=bias,
                         batch_size=batch_size, batch_norm=batch_norm,
                         param_specific_learn_rate=param_specific_learn_rate,
                         save=save, on_gpu=on_gpu, name=name,
                         standardize_X=standardize_X, standardize_y=standardize_y,
                         **kwargs)

    def init_network(self, **kwargs):
        """
        Set up the network structure by creating the Layer objects and
        connecting them together. The second layer is a DAS_Layer.
        This subroutine overwrites a subroutine from the parent ANN class.

        Returns
        -------
        None.

        """

        self.layers = []

        n_hidden = self.n_layers

        # add the input layer
        self.layers.append(Layer(self.n_in, 0, self.n_layers, 'linear',
                                 self.loss, False, batch_size=self.batch_size,
                                 batch_norm = False,
                                 lamb=self.lamb, on_gpu=self.on_gpu))

        # by default, the 1st layer does not have a bias neuron. This way
        # the orthogonal vectors are only related to the D inputs, and not the
        # D+1 inputs + 1 bias neuron, which does not make physical sense
        self.bias[0] = False

        # add the deep active subspace layer
        self.layers.append(
            DAS_Layer(
                self.d,
                self.n_layers,
                self.bias[1],
                activation=self.activation_das,
                batch_size=self.batch_size))

        # adjust layer_activation of the ANN superclass
        self.layer_activation[1] = self.activation_das

        # add the hidden layers
        for r in range(2, n_hidden):
            self.layers.append(Layer(self.n_neurons, r, self.n_layers, self.layer_activation[r],
                                     self.loss, self.bias[r], batch_size=self.batch_size,
                                     batch_norm=self.batch_norm[r],
                                     lamb=self.lamb, on_gpu=self.on_gpu))

        # add the output layer
        self.layers.append(
            Layer(
                self.n_out,
                r + 1,
                self.n_layers,
                self.activation_out,
                self.loss,
                batch_size=self.batch_size,
                batch_norm=False,
                lamb=self.lamb,
                n_softmax=self.n_softmax,
                on_gpu=self.on_gpu,
                **kwargs))



"""
==============================================================================
CLASS FOR A QUANTIZED SOFTMAX NETWORK
------------------------------------------------------------------------------
Author: W. Edeling
Source: Resampling with neural networks for stochastic parameterization in
        multiscale systems
arXiv preprint arXiv:2004.01457
==============================================================================
"""
from itertools import chain
import numpy as np
# import h5py
from ..campaign import Campaign
import easysurrogate as es


class QSN_Surrogate(Campaign):

    def __init__(self, **kwargs):
        print('Creating Quantized Softmax Object')
        self.name = 'QSN Surrogate'
        # create a Feature_Engineering object
        self.feat_eng = es.methods.Feature_Engineering()

    ############################
    # START COMMON SUBROUTINES #
    ############################

    def train(self, feats, target, n_iter, lags=None, local=False,
              n_bins=10, test_frac=0.0,
              n_layers=2, n_neurons=100,
              activation='leaky_relu',
              batch_size=64, lamb=0.0,
              standardize_X=True, **kwargs):
        """
        Perform back propagation to train the QSN

        Parameters
        ----------
        feats : feature array, or list of different feature arrays
        target : the target data
        lags : list of time lags per feature
        n_iter : number of back propagation iterations
        n_bins : Number of output bins. The default is 10.
        test_frac : Fraction of the data to use for training.
                    The default is 0.0.
        n_layers : The number of layers in the neural network. Includes the
                   input layer and the hidden layers. The default is 2.
        n_neurons : The number of neurons per layer. The default is 100.
        activation : Type of activation function. The default is 'leaky_relu'.
        batch_size : Mini batch size. The default is 64.
        lamb : L2 regularization parameter. The default is 0.0.
        standardize_X : standardize the input features. Default is True.

        Returns
        -------
        None.

        """

        # is a single array is provided, also put it in a list
        if isinstance(feats, np.ndarray):
            feats = [feats]

        # time lags
        self.lags = lags

        self.n_bins = n_bins

        # flag if the surrogate is to be applied locally or not
        self.local = local

        # test fraction
        self.test_frac = test_frac

        # prepare the training data
        X_train, y_train, _, _ = self.feat_eng.get_training_data(
            feats, target, lags=lags, local=local, test_frac=test_frac)

        # get the maximum lag that was specified
        self.max_lag = self.feat_eng.max_lag

        # number of softmax layers (one per output)
        self.n_softmax = y_train.shape[1]

        # number of output neurons
        n_out = n_bins * self.n_softmax

        # one-hot encoded y data e.g. [0, 0, 1, 0, ..., 0] if the y sample
        # falls in the 3rd bin
        one_hot_encoded_data = self.feat_eng.bin_data(y_train, n_bins)

        # simple sampler to draw random samples from the bins
        self.sampler = es.methods.SimpleBin(self.feat_eng)

        # create the feed-forward QSN
        self.neural_net = es.methods.ANN(X=X_train, y=one_hot_encoded_data,
                                         n_layers=n_layers, n_neurons=n_neurons,
                                         n_softmax=self.n_softmax, n_out=n_out,
                                         loss='cross_entropy',
                                         activation=activation, batch_size=batch_size,
                                         lamb=lamb, decay_step=10**4, decay_rate=0.9,
                                         standardize_X=standardize_X, standardize_y=False,
                                         save=False)

        print('===============================')
        print('Training Quantized Softmax Network...')

        # train network for N_iter mini batches
        self.neural_net.train(n_iter, store_loss=True)
        self.set_data_stats()
        if lags is not None:
            self.feat_eng.initial_condition_feature_history(feats)

    def predict(self, X):
        """
        Make a stochastic prediction of the output y conditional on the
        input features [X_t, X_{t-lag1}, X_{t-lag2}, ...]

        Parameters
        ----------
        X: the state at the current (time) step. If the QSN is conditioned on
        more than 1 (time-lagged) variable, X must be a list containing all
        variables at the current time step.

        Returns
        -------
        Stochastic prediction of the output y

        """

        return self.feat_eng._predict(X, self._feed_forward)

    def _feed_forward(self, feat):
        """

        A feed forward run of the QSN. This is the only part of prediction that is specific
        to QSN_Surrogate.

        Parameters
        ----------
        feat : array of list of arrays
               The feature array of a list of feature arrays on which to evaluate the surrogate.

        Returns
        -------
        y : array
            the stochastic prediction of the QSN.
        """

        # features were standardized during training, do so here as well
        feat = (feat - self.feat_mean) / self.feat_std
        # o_i = the probability mass function at the output layer
        # max_idx = the bin index with the highest probability
        o_i, max_idx, _ = self.neural_net.get_softmax(feat.reshape([1, self.neural_net.n_in]))
        # resample a value from the selected bin
        return self.sampler.resample(max_idx)

    def save_state(self):
        """
        Save the state of the QSN surrogate to a pickle file
        """
        state = self.__dict__
        super().save_state(state=state, name=self.name)

    def load_state(self):
        """
        Load the state of the QSN surrogate from file
        """
        super().load_state(name=self.name)

    ##########################
    # END COMMON SUBROUTINES #
    ##########################

    def set_data_stats(self):
        """
        If the data were standardized, this stores the mean and
        standard deviation of the features in the ANN_Surrogate object.
        """
        if hasattr(self.neural_net, 'X_mean'):
            self.feat_mean = self.neural_net.X_mean
            self.feat_std = self.neural_net.X_std
        else:
            self.feat_mean = 0.0
            self.feat_std = 1.0

        if hasattr(self.neural_net, 'y_mean'):
            self.output_mean = self.neural_net.y_mean
            self.output_std = self.neural_net.y_std
        else:
            self.output_mean = 0.0
            self.output_std = 1.0

    def get_dimensions(self):
        """
        Get some useful dimensions of the QSN surrogate. Returns a dict with the number
        of training samples (n_train), the number of data samples (n_samples),
        the number of test samples (n_test), the number of input neurons (n_in),
        and the number of output_neurons (n_out).

        Returns
        -------
        dims : dict
            The dimensions dictionary.

        """

        dims = {}
        dims['n_train'] = self.feat_eng.n_train
        dims['n_samples'] = self.feat_eng.n_samples
        dims['n_test'] = dims['n_samples'] - dims['n_train']
        dims['n_in'] = self.neural_net.n_in
        dims['n_out'] = self.neural_net.n_out

        return dims



from .QSN_analysis import QSN_analysis
from .CCM_analysis import CCM_analysis
from .KMN_analysis import KMN_analysis
from .ANN_analysis import ANN_analysis
from .DAS_analysis import DAS_analysis
from .base import BaseAnalysis
from .GP_analysis import GP_analysis
